# Copyright (C) 2008 Zed A. Shaw.  Licensed under the terms of the GPLv3.
from __future__ import print_function

import socket
import sys

from mock import Mock, patch
import lmtpd
import six

from nose.tools import assert_equal, assert_raises, with_setup
from salmon import mail, queue, routing, server

from .message_tests import (test_mail_request, test_mail_response_attachments, test_mail_response_html,
                            test_mail_response_html_and_plain_text, test_mail_response_plain_text)
from .setup_env import setup_salmon_dirs, teardown_salmon_dirs

SMTP_MESSAGE_DEFS = {
    2: {"ok": u"250 Ok\r\n".encode()},
    3: {"ok": u"250 OK\r\n".encode()},
}

SMTP_MESSAGES = SMTP_MESSAGE_DEFS[sys.version_info[0]]


def test_router():
    routing.Router.deliver(test_mail_request())

    # test that fallthrough works too
    msg = test_mail_request()
    msg['to'] = 'unhandled@localhost'
    msg.To = msg['to']

    routing.Router.deliver(msg)


@patch("asynchat.async_chat.push")
def test_SMTPChannel(push_mock):
    channel = server.SMTPChannel(Mock(), Mock(), Mock())
    expected_version = u"220 {} {}\r\n".format(socket.getfqdn(), server.smtpd.__version__).encode()

    assert_equal(push_mock.call_args[0][1:], (expected_version,))
    assert_equal(type(push_mock.call_args[0][1]), six.binary_type)

    push_mock.reset_mock()
    channel.seen_greeting = True
    channel.smtp_MAIL("FROM: you@example.com\r\n")
    assert_equal(push_mock.call_args[0][1:], (SMTP_MESSAGES["ok"],))

    push_mock.reset_mock()
    channel.smtp_RCPT("TO: me@example.com")
    assert_equal(push_mock.call_args[0][1:], (SMTP_MESSAGES["ok"],))

    push_mock.reset_mock()
    channel.smtp_RCPT("TO: them@example.com")
    assert_equal(push_mock.call_args[0][1:],
                 (u"451 Will not accept multiple recipients in one transaction\r\n".encode(),))


def test_SMTPReceiver_process_message():
    receiver = server.SMTPReceiver(host="localhost", port=0)
    msg = test_mail_request()

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg))
        assert response is None, response

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        router_mock.deliver.side_effect = Exception()
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg))
        assert response is None, response

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        router_mock.deliver.side_effect = server.SMTPError(450, "Not found")
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg))
        assert_equal(response, "450 Not found")

    # Python 3's smtpd takes some extra kawrgs, but i we don't care about that at the moment
    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg), mail_options=[], rcpt_options=[])
        assert response is None, response


@patch("lmtpd.asynchat.async_chat.push")
def test_LMTPChannel(push_mock):
    channel = lmtpd.LMTPChannel(Mock(), Mock(), Mock())
    expected_version = u"220 {} {}\r\n".format(socket.getfqdn(), server.lmtpd.__version__).encode()

    assert_equal(push_mock.call_args[0][1:], (expected_version,))
    assert_equal(type(push_mock.call_args[0][1]), six.binary_type)

    push_mock.reset_mock()
    channel.seen_greeting = True
    channel.lmtp_MAIL(b"FROM: you@example.com\r\n")
    assert_equal(push_mock.call_args[0][1:], (u"250 2.1.0 Ok\r\n".encode(),))

    push_mock.reset_mock()
    channel.lmtp_RCPT(b"TO: me@example.com")
    assert_equal(push_mock.call_args[0][1:], (u"250 2.1.0 Ok\r\n".encode(),))

    push_mock.reset_mock()
    channel.lmtp_RCPT(b"TO: them@example.com")
    assert_equal(push_mock.call_args[0][1:], (u"250 2.1.0 Ok\r\n".encode(),))


def test_LMTPReceiver_process_message():
    receiver = server.LMTPReceiver(host="localhost", port=0)
    msg = test_mail_request()

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg))
        assert response is None, response

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        router_mock.deliver.side_effect = Exception()
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg))
        assert response is None, response

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        router_mock.deliver.side_effect = server.SMTPError(450, "Not found")
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg))
        assert_equal(response, "450 Not found")

    # lmtpd's server is a subclass of smtpd's server, so we should support the same kwargs here
    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        response = receiver.process_message(msg.Peer, msg.From, msg.To, str(msg), mail_options=[], rcpt_options=[])
        assert response is None, response


@patch("salmon.queue.Queue")
def test_QueueReceiver_process_message(queue_mock):
    receiver = server.QueueReceiver("run/queue/thingy")
    msg = test_mail_request()

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        response = receiver.process_message(msg)
        assert response is None, response

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        router_mock.deliver.side_effect = Exception()
        response = receiver.process_message(msg)
        assert response is None, response

    with patch("salmon.server.routing.Router") as router_mock, \
            patch("salmon.server.undeliverable_message"):
        router_mock.deliver.side_effect = server.SMTPError(450, "Not found")
        response = receiver.process_message(msg)
        # queue doesn't actually support SMTPErrors
        assert response is None, response


def test_Relay_asserts_ssl_options():
    """Relay raises an AssertionError if the ssl option is used in combination with starttls or lmtp"""
    with assert_raises(AssertionError):
        server.Relay("localhost", ssl=True, starttls=True)

    with assert_raises(AssertionError):
        server.Relay("localhost", ssl=True, lmtp=True)

    with assert_raises(AssertionError):
        server.Relay("localhost", ssl=True, starttls=True, lmtp=True)

    # no error
    server.Relay("localhost", starttls=True, lmtp=True)


@patch("salmon.server.smtplib.SMTP")
def test_relay_deliver(client_mock):
    # test that relay will actually call smtplib.SMTP
    relay = server.Relay("localhost", port=0)

    relay.deliver(test_mail_response_plain_text())
    assert_equal(client_mock.return_value.sendmail.call_count, 1)

    relay.deliver(test_mail_response_html())
    assert_equal(client_mock.return_value.sendmail.call_count, 2)

    relay.deliver(test_mail_response_html_and_plain_text())
    assert_equal(client_mock.return_value.sendmail.call_count, 3)

    relay.deliver(test_mail_response_attachments())
    assert_equal(client_mock.return_value.sendmail.call_count, 4)


@patch("salmon.server.smtplib.SMTP")
def test_relay_smtp(client_mock):
    relay = server.Relay("localhost", port=0)
    relay.deliver(test_mail_response_plain_text())
    assert_equal(client_mock.return_value.sendmail.call_count, 1)
    assert_equal(client_mock.return_value.starttls.call_count, 0)

    client_mock.reset_mock()
    relay = server.Relay("localhost", port=0, starttls=True)
    relay.deliver(test_mail_response_plain_text())
    assert_equal(client_mock.return_value.sendmail.call_count, 1)
    assert_equal(client_mock.return_value.starttls.call_count, 1)


@patch("salmon.server.smtplib.LMTP")
def test_relay_lmtp(client_mock):
    relay = server.Relay("localhost", port=0, lmtp=True)
    relay.deliver(test_mail_response_plain_text())
    assert_equal(client_mock.return_value.sendmail.call_count, 1)


@patch("salmon.server.smtplib.SMTP_SSL")
def test_relay_smtp_ssl(client_mock):
    relay = server.Relay("localhost", port=0, ssl=True)
    relay.deliver(test_mail_response_plain_text())
    assert_equal(client_mock.return_value.sendmail.call_count, 1)


@patch('salmon.server.resolver.query')
@patch("salmon.server.smtplib.SMTP")
def test_relay_deliver_mx_hosts(client_mock, query):
    query.return_value = [Mock()]
    query.return_value[0].exchange = "localhost"
    relay = server.Relay(None, port=0)

    msg = test_mail_response_plain_text()
    msg['to'] = 'user@localhost'
    relay.deliver(msg)
    assert_equal(query.call_count, 1)


@patch('salmon.server.resolver.query')
def test_relay_resolve_relay_host(query):
    from dns import resolver
    query.side_effect = resolver.NoAnswer
    relay = server.Relay(None, port=0)
    host = relay.resolve_relay_host('user@localhost')
    assert_equal(host, 'localhost')
    assert_equal(query.call_count, 1)

    query.reset_mock()
    query.side_effect = None  # reset_mock doens't clear return_value or side_effect
    query.return_value = [Mock()]
    query.return_value[0].exchange = "mx.example.com"
    host = relay.resolve_relay_host('user@example.com')
    assert_equal(host, 'mx.example.com')
    assert_equal(query.call_count, 1)


@patch("salmon.server.smtplib.SMTP")
def test_relay_reply(client_mock):
    relay = server.Relay("localhost", port=0)
    print("Relay: %r" % relay)

    relay.reply(test_mail_request(), 'from@localhost', 'Test subject', 'Body')
    assert_equal(client_mock.return_value.sendmail.call_count, 1)


@patch("socket.create_connection")
def test_relay_raises_exception(create_mock):
    # previously, salmon would eat up socket errors and just log something. Not cool!
    create_mock.side_effect = socket.error
    relay = server.Relay("example.com", port=0)
    with assert_raises(socket.error):
        relay.deliver(test_mail_response_plain_text())


def raises_exception(*x, **kw):
    raise RuntimeError("Raised on purpose.")


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('salmon.routing.Router', new=Mock())
def test_queue_receiver():
    receiver = server.QueueReceiver('run/queue')
    run_queue = queue.Queue('run/queue')
    run_queue.push(str(test_mail_response_plain_text()))
    assert run_queue.count() > 0
    receiver.start(one_shot=True)
    assert_equal(run_queue.count(), 0)

    routing.Router.deliver.side_effect = raises_exception
    receiver.process_message(mail.MailRequest('localhost', 'test@localhost', 'test@localhost', 'Fake body.'))


@patch('threading.Thread', new=Mock())
@patch('salmon.routing.Router', new=Mock())
def test_SMTPReceiver():
    receiver = server.SMTPReceiver(port=0)
    receiver.start()
    receiver.process_message('localhost', 'test@localhost', 'test@localhost',
                             'Fake body.')

    routing.Router.deliver.side_effect = raises_exception
    receiver.process_message('localhost', 'test@localhost', 'test@localhost', 'Fake body.')

    receiver.close()


def test_SMTPError():
    err = server.SMTPError(550)
    assert_equal(str(err), '550 Permanent Failure Mail Delivery Protocol Status')

    err = server.SMTPError(400)
    assert_equal(str(err), '400 Persistent Transient Failure Other or Undefined Status')

    err = server.SMTPError(425)
    assert_equal(str(err), '425 Persistent Transient Failure Mailbox Status')

    err = server.SMTPError(999)
    assert_equal(str(err), "999 ")

    err = server.SMTPError(999, "Bogus Error Code")
    assert_equal(str(err), "999 Bogus Error Code")



import os
import shutil

import config.testing  # noqa

dirs = ["run", "logs"]


def setup_package():
    for path in dirs:
        os.mkdir(path)


def teardown_package():
    for path in dirs:
        shutil.rmtree(path, ignore_errors=True)



from mock import Mock, patch

from nose.tools import assert_equal, assert_raises, raises, with_setup
from salmon import routing
from salmon.mail import MailRequest
from salmon.routing import MemoryStorage, Router, ShelveStorage, StateStorage, route

from .setup_env import setup_router, setup_salmon_dirs, teardown_salmon_dirs


def test_MemoryStorage():
    store = MemoryStorage()
    store.set(test_MemoryStorage.__module__, "tester@localhost", "TESTED")

    assert_equal(store.get(test_MemoryStorage.__module__, "tester@localhost"), "TESTED")

    assert_equal(store.get(test_MemoryStorage.__module__, "tester2@localhost"), "START")

    store.clear()

    assert_equal(store.get(test_MemoryStorage.__module__, "tester@localhost"), "START")


def test_ShelveStorage():
    store = ShelveStorage("run/states.db")

    store.set(test_ShelveStorage.__module__, "tester@localhost", "TESTED")
    assert_equal(store.get(test_MemoryStorage.__module__, "tester@localhost"), "TESTED")

    assert_equal(store.get(test_MemoryStorage.__module__, "tester2@localhost"), "START")

    store.clear()
    assert_equal(store.get(test_MemoryStorage.__module__, "tester@localhost"), "START")


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_RoutingBase():
    # check that Router is in a pristine state
    assert_equal(len(Router.ORDER), 0)
    assert_equal(len(Router.REGISTERED), 0)

    setup_router(['salmon_tests.handlers.simple_fsm_mod'])
    from .handlers import simple_fsm_mod

    assert_equal(len(Router.ORDER), 5)
    assert_equal(len(Router.REGISTERED), 5)

    message = MailRequest('fakepeer', 'zedshaw@localhost', 'users-subscribe@localhost', "")
    Router.deliver(message)
    assert Router.in_state(simple_fsm_mod.CONFIRM, message)

    confirm = MailRequest('fakepeer', '"Zed Shaw" <zedshaw@localhost>',  'users-confirm-1@localhost', "")
    Router.deliver(confirm)
    assert Router.in_state(simple_fsm_mod.POSTING, message)

    Router.deliver(message)
    assert Router.in_state(simple_fsm_mod.NEXT, message)

    Router.deliver(message)
    assert Router.in_state(simple_fsm_mod.END, message)

    Router.deliver(message)
    assert Router.in_state(simple_fsm_mod.START, message)

    Router.clear_states()
    Router.LOG_EXCEPTIONS = True
    explosion = MailRequest('fakepeer', '<hacker@localhost>', 'start-explode@localhost', "")
    Router.deliver(explosion)

    assert Router.in_error(simple_fsm_mod.END, explosion)

    Router.clear_states()
    Router.LOG_EXCEPTIONS = False
    explosion = MailRequest('fakepeer',  'hacker@localhost', 'start-explode@localhost', "")
    assert_raises(RuntimeError, Router.deliver, explosion)

    Router.reload()
    assert 'salmon_tests.handlers.simple_fsm_mod' in Router.HANDLERS
    assert_equal(len(Router.ORDER), 5)
    assert_equal(len(Router.REGISTERED), 5)


def test_Router_undeliverable_queue():
    Router.clear_routes()
    Router.clear_states()

    Router.UNDELIVERABLE_QUEUE = Mock()
    msg = MailRequest('fakepeer', 'from@localhost', 'to@localhost', "Nothing")

    Router.deliver(msg)
    assert_equal(Router.UNDELIVERABLE_QUEUE.push.call_count, 1)


@raises(NotImplementedError)
def test_StateStorage_get_raises():
    s = StateStorage()
    s.get("raises", "raises")


@raises(NotImplementedError)
def test_StateStorage_set_raises():
    s = StateStorage()
    s.set("raises", "raises", "raises")


@raises(NotImplementedError)
def test_StateStorage_clear_raises():
    s = StateStorage()
    s.clear()


@raises(TypeError)
def test_route___get___raises():
    class BadRoute(object):

        @route("test")
        def wont_work(message, **kw):
            pass

    br = BadRoute()
    br.wont_work("raises")


@patch('salmon.routing.reload', new=Mock(side_effect=ImportError))
@patch('salmon.routing.LOG', new=Mock())
def test_reload_raises():
    Router.LOG_EXCEPTIONS = True
    Router.reload()
    assert_equal(routing.LOG.exception.call_count, 1)

    Router.LOG_EXCEPTIONS = False
    routing.LOG.exception.reset_mock()
    assert_raises(ImportError, Router.reload)
    assert_equal(routing.LOG.exception.call_count, 0)

    routing.LOG.exception.reset_mock()
    Router.LOG_EXCEPTIONS = True
    Router.load(['fake.handler'])
    assert_equal(routing.LOG.exception.call_count, 1)

    Router.LOG_EXCEPTIONS = False
    routing.LOG.exception.reset_mock()
    assert_raises(ImportError, Router.load, ['fake.handler'])
    assert_equal(routing.LOG.exception.call_count, 0)



from nose.tools import assert_equal
from salmon import view


def test_load():
    template = view.load("template.txt")
    assert template
    assert template.render()


def test_render():
    # try with some empty vars
    text = view.render({}, "template.txt")
    assert text


def test_most_basic_form():
    msg = view.respond(locals(), 'template.txt')
    assert msg.Body


def test_respond_cadillac_version():
    person = 'Tester'

    msg = view.respond(locals(), Body='template.txt', Html='template.html', From='test@localhost',
                       To='receiver@localhost', Subject='Test body from "%(person)s".')

    assert msg.Body
    assert msg.Html

    for k in ['From', 'To', 'Subject']:
        assert k in msg


def test_respond_plain_text():
    person = 'Tester'

    msg = view.respond(locals(), Body='template.txt', From='test@localhost', To='receiver@localhost',
                       Subject='Test body from "%(person)s".')

    assert msg.Body
    assert not msg.Html

    for k in ['From', 'To', 'Subject']:
        assert k in msg


def test_respond_html_only():
    person = 'Tester'

    msg = view.respond(locals(), Html='template.html', From='test@localhost', To='receiver@localhost',
                       Subject='Test body from "%(person)s".')

    assert not msg.Body
    assert msg.Html

    for k in ['From', 'To', 'Subject']:
        assert k in msg


def test_respond_attach():
    person = "hello"
    mail = view.respond(locals(), Body="template.txt", From="test@localhost", To="receiver@localhost",
                        Subject='Test body from someone.')

    view.attach(mail, locals(), 'template.html', content_type="text/html", filename="template.html",
                disposition='attachment')

    assert_equal(len(mail.attachments), 1)

    msg = mail.to_message()
    assert_equal(len(msg.get_payload()), 2)
    assert str(msg)

    mail.clear()

    view.attach(mail, locals(), 'template.html', content_type="text/html")
    assert_equal(len(mail.attachments), 1)

    msg = mail.to_message()
    assert_equal(len(msg.get_payload()), 2)
    assert str(msg)


def test_unicode():
    person = u'H\xe9avy M\xe9t\xe5l Un\xeec\xf8d\xe9'
    mail = view.respond(locals(), Html="unicode.html", From="test@localhost", To="receiver@localhost",
                        Subject='Test body from someone.')

    assert str(mail)

    view.attach(mail, locals(), "unicode.html", filename="attached.html")

    assert str(mail)



# -*- coding: utf-8 -*-
"""
Salmon takes the policy that email it receives is most likely complete garbage
using bizarre pre-Unicode formats that are irrelevant and unnecessary in
today's modern world. These are turned into something nice and clean that a
regular Python programmer can work with:  Unicode.

That's the receiving end, but on the sending end Salmon wants to make the world
better by not increasing the suffering. To that end, Salmon will canonicalize
all email it sends to be ascii or utf-8 (whichever is simpler and works to
encode the data). It is possible to use other encodings (Salmon doesn't live in
some fictional world), but this generally frowned upon.

To accomplish these tasks, Salmon goes back to basics and assert a few simple
rules on each email it receives:

1) NO ENCODING IS TRUSTED, NO LANGUAGE IS SACRED, ALL ARE SUSPECT.
2) Python wants Unicode, it will get Unicode.
3) Any email that CANNOT become Unicode, CANNOT be processed by Salmon or
   Python.
4) Email addresses are ESSENTIAL to Salmon's routing and security, and therefore
   will be canonicalized and properly encoded.
5) Salmon will therefore try to "upgrade" all email it receives to Unicode
   internally, and cleaning all email addresses.
6) It does this by decoding all codecs, and if the codec LIES, then it will
   attempt to statistically detect the codec using chardet.
7) If it can't detect the codec, and the codec lies, then the email is bad.
8) All text bodies and attachments are then converted to Python unicode/str
   (for Python 2.7 and 3.x respectively) in the same way as the headers.
9) All other attachments are converted to raw strings as-is.

Once Salmon has done this, your Python handler can now assume that all
MailRequest objects are happily Unicode enabled and ready to go. The rule is:

    IF IT CANNOT BE UNICODE, THEN PYTHON CANNOT WORK WITH IT.

On the outgoing end (when you send a MailResponse), Salmon tries to create the
email it wants to receive by canonicalizing it:

1) All email will be encoded in the simplest cleanest way possible without
   losing information.
2) All headers are converted to 'ascii', and if that doesn't work, then
   'utf-8'.
3) All text/* attachments and bodies are converted to ascii, and if that
   doesn't work, 'utf-8'. It is possible to override this, but you're a bad
   person if you do
4) All other attachments are left alone.
5) All email addresses are normalized and encoded if they have not been
   already.

The end result is an email that has the highest probability of not containing
any obfuscation techniques, hidden characters, bad characters, improper
formatting, invalid non-characterset headers, or any of the other billions of
things email clients do to the world. The output rule of Salmon is:

    ALL EMAIL IS ASCII FIRST, THEN ENCODED ASCII-SAFE, AND IF IT CANNOT BE
    EITHER OF THOSE IT WILL NOT BE SENT.

Following these simple rules, this module does the work of converting email to
the canonical format and sending the canonical format. The code is probably the
most complex part of Salmon since the job it does is difficult.

Test results show that Salmon can safely canonicalize most email from any
culture (not just English) to the canonical form, and that if it can't then the
email is not formatted right and/or spam.

If you find an instance where this is not the case, then submit it to the
project as a test case.
"""
from __future__ import print_function, unicode_literals

from email import encoders
from email.charset import Charset
from email.message import Message
from email.utils import parseaddr
import email
import re
import string
import warnings

import chardet
import six

DEFAULT_ENCODING = "utf-8"
DEFAULT_ERROR_HANDLING = "strict"
CONTENT_ENCODING_KEYS = set(['Content-Type', 'Content-Transfer-Encoding',
                             'Content-Disposition', 'Mime-Version'])
CONTENT_ENCODING_REMOVED_PARAMS = ['boundary']

REGEX_OPTS = re.IGNORECASE | re.MULTILINE
ENCODING_REGEX = re.compile(r"\=\?([a-z0-9\-]+?)\?([bq])\?", REGEX_OPTS)
ENCODING_END_REGEX = re.compile(r"\?=", REGEX_OPTS)
INDENT_REGEX = re.compile(r"\n\s+")

ADDRESS_HEADERS_WHITELIST = ['From', 'To', 'Delivered-To', 'Cc', 'Bcc']


def VALUE_IS_EMAIL_ADDRESS(v):
    return "@" in v


class EncodingError(Exception):
    """Thrown when there is an encoding error."""
    pass


class ContentEncoding(object):
    """
    Wrapper various content encoding headers

    The value of each key is returned as a tuple of a string and a dict of
    params. Note that changes to the params dict won't be reflected in the
    underlying MailBase unless the tuple is reassigned:

        >>> value = mail.content_encoding["Content-Type"]
        >>> print(value)
        ('text/html', {'charset': 'us-ascii'})
        >>> value[1]['charset'] = 'utf-8'
        >>> print(mail["Content-Type"])  # unchanged
        ('text/html', {'charset': 'us-ascii'})
        >>> mail.content_encoding["Content-Type"] = value
        >>> print(mail["Content-Type"])
        ('text/html', {'charset': 'utf-8'})

    Will raise EncodingError if you try to access a header that isn't in
    ``CONTENT_ENCODING_KEYS``
    """
    def __init__(self, base):
        self.base = base
        self.defaults = {
            "Content-Transfer-Encoding": ("7bit", {}),
        }

    def get(self, key, default=None):
        if key not in CONTENT_ENCODING_KEYS:
            raise EncodingError("EncodingError: %s is not in CONTENT_ENCODING_KEYS" % key)

        value, params = parse_parameter_header(self.base.mime_part, key)
        value = value.lower() if value else value
        return (value, params)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        if key not in CONTENT_ENCODING_KEYS:
            raise EncodingError("EncodingError: %s is not in CONTENT_ENCODING_KEYS" % key)

        # remove any other content header before adding our own
        del self.base.mime_part[key]
        self.base.mime_part.add_header(key, value[0], **value[1])

    def __delitem__(self, key):
        if key not in CONTENT_ENCODING_KEYS:
            raise EncodingError("EncodingError: %s is not in CONTENT_ENCODING_KEYS" % key)

        del self.base.mime_part[key]

    def __len__(self):
        return len(CONTENT_ENCODING_KEYS)

    def __contains__(self, key):
        return key in CONTENT_ENCODING_KEYS

    def keys(self):
        return CONTENT_ENCODING_KEYS


class MailBase(object):
    """
    MailBase is used as the basis of salmon.mail and contains the basics of
    encoding an email.  You actually can do all your email processing with this
    class, but it's more raw.
    """
    def __init__(self, mime_part_or_headers=None, parent=None):
        self.parts = []
        self.parent = parent
        self.content_encoding = ContentEncoding(self)

        if isinstance(mime_part_or_headers, Message):
            self.mime_part = mime_part_or_headers
        else:
            self.mime_part = Message()
            if mime_part_or_headers is not None and len(mime_part_or_headers) > 0:
                for key, value in mime_part_or_headers:
                    self.mime_part[key] = value

    def __getitem__(self, key):
        header = self.mime_part.get(key)
        return header_from_mime_encoding(header)

    def __len__(self):
        return len(self.mime_part)

    def __iter__(self):
        for k in self.mime_part.keys():
            yield k

    def __contains__(self, key):
        return key in self.mime_part

    def __setitem__(self, key, value):
        del self.mime_part[key]
        self.mime_part[normalize_header(key)] = value

    def __delitem__(self, key):
        del self.mime_part[key]

    def __nonzero__(self):
        return self.body is not None or len(self.mime_part) > 0 or len(self.parts) > 0

    def items(self):
        return [(normalize_header(key), header_from_mime_encoding(header)) for key, header in self.mime_part.items()]

    def keys(self):
        """Returns header keys."""
        return [normalize_header(key) for key in self.mime_part.keys()]

    def append_header(self, key, value):
        """Like __set_item__, but won't replace header values"""
        self.mime_part[normalize_header(key)] = value

    def get_all(self, key):
        return self.mime_part.get_all(key, [])

    @property
    def body(self):
        body = self.mime_part.get_payload(decode=True)
        if body:
            # decode the payload according to the charset given if it's text
            ctype, params = self.content_encoding['Content-Type']

            if not ctype:
                charset = 'ascii'
                body = attempt_decoding(charset, body)
            elif ctype.startswith("text/"):
                charset = params.get('charset', 'ascii')
                body = attempt_decoding(charset, body)
            else:
                # it's a binary codec of some kind, so just decode and leave it
                # alone for now
                pass
        return body

    @body.setter
    def body(self, value):
        ctype, params = self.content_encoding['Content-Type']
        self.mime_part.set_payload(value, params.get("charset", None))

    def attach_file(self, filename, data, ctype, disposition):
        """
        A file attachment is a raw attachment with a disposition that
        indicates the file name.
        """
        assert filename, "You can't attach a file without a filename."
        assert ctype.lower() == ctype, "Hey, don't be an ass.  Use a lowercase content type."

        part = MailBase(parent=self)
        part.body = data
        part.content_encoding['Content-Type'] = (ctype, {'name': filename})
        part.content_encoding['Content-Disposition'] = (disposition,
                                                        {'filename': filename})
        self.parts.append(part)

    def attach_text(self, data, ctype):
        """
        This attaches a simpler text encoded part, which doesn't have a
        filename.
        """
        assert ctype.lower() == ctype, "Hey, don't be an ass.  Use a lowercase content type."

        part = MailBase(parent=self)
        part.body = data
        part.content_encoding['Content-Type'] = (ctype, {})
        self.parts.append(part)

    def walk(self):
        for p in self.parts:
            yield p
            for x in p.walk():
                yield x


class MIMEPart(Message):
    """
    A reimplementation of nearly everything in email.mime to be more useful
    for actually attaching things.  Rather than one class for every type of
    thing you'd encode, there's just this one, and it figures out how to
    encode what you ask it.
    """
    def __init__(self, type_, **params):
        self.mimetype = type_

        # classes from email.* are all old-style in Python, so don't replace
        # this with super()
        Message.__init__(self)

        self.add_header('Content-Type', type_, **params)

    def add_text(self, content, charset=None):
        # this is text, so encode it in canonical form
        if charset is not None:
            warnings.warn("You are adding text that is neither ASCII nor UTF-8. Please reconsider your choice.",
                          UnicodeWarning)

        charset = charset or 'ascii'
        try:
            encoded = content.encode(charset)
        except UnicodeError:
            encoded = content.encode('utf-8')
            charset = 'utf-8'
        except AttributeError:
            # content is already bytes
            encoded = content

        self.set_payload(encoded, charset=charset)

    def extract_payload(self, mail):
        if mail.body is None:
            return  # only None, '' is still ok

        ctype, ctype_params = mail.content_encoding['Content-Type']
        cdisp, cdisp_params = mail.content_encoding['Content-Disposition']

        assert ctype, "Extract payload requires that mail.content_encoding have a valid Content-Type."

        if ctype.startswith("text/"):
            self.add_text(mail.body, charset=ctype_params.get('charset'))
        else:
            if cdisp:
                # replicate the content-disposition settings
                self.add_header('Content-Disposition', cdisp, **cdisp_params)

            self.set_payload(mail.body)
            encoders.encode_base64(self)

    def __repr__(self):
        return "<MIMEPart '%s': %r, %r, multipart=%r>" % (self.mimetype, self['Content-Type'],
                                                          self['Content-Disposition'],
                                                          self.is_multipart())


def from_message(message, parent=None):
    """
    Given a MIMEBase or similar Python email API message object, this
    will canonicalize it and give you back a pristine MailBase.
    If it can't then it raises a EncodingError.
    """
    mail = MailBase(message, parent)

    if message.is_multipart():
        # recursively go through each subpart and decode in the same way
        for msg in message.get_payload():
            if msg != message:  # skip the multipart message itself
                mail.parts.append(from_message(msg, mail))

    return mail


def to_message(mail):
    """
    Given a MailBase message, this will construct a MIMEPart
    that is canonicalized for use with the Python email API.

    N.B. this changes the original email.message.Message
    """
    ctype, params = mail.content_encoding['Content-Type']
    if not ctype:
        if mail.parts:
            ctype = 'multipart/mixed'
        else:
            ctype = 'text/plain'
    else:
        if mail.parts:
            assert ctype.startswith("multipart") or ctype.startswith("message"), \
                    "Content type should be multipart or message, not %r" % ctype

    # adjust the content type according to what it should be now
    mail.content_encoding['Content-Type'] = (ctype, params)

    try:
        out = MIMEPart(ctype, **params)
    except TypeError as exc:
        raise EncodingError("Content-Type malformed, not allowed: %r; %r (Python ERROR: %s" %
                            (ctype, params, getattr(exc, "message", "(No error message)")))

    for k in mail.keys():
        if k in ADDRESS_HEADERS_WHITELIST:
            value = header_to_mime_encoding(mail[k])
        else:
            value = header_to_mime_encoding(mail[k], not_email=True)

        if k.lower() in [key.lower() for key in CONTENT_ENCODING_KEYS]:
            del out[k]
            out[k] = value
        else:
            out[k] = value

    out.extract_payload(mail)

    # make sure payload respects cte
    cte, cte_params = mail.content_encoding['Content-Transfer-Encoding']
    if cte == "quoted-printable":
        del out['Content-Transfer-Encoding']
        encoders.encode_quopri(out)
    elif cte == "base64":
        del out['Content-Transfer-Encoding']
        encoders.encode_base64(out)

    # go through the children
    for part in mail.parts:
        out.attach(to_message(part))

    return out


def to_string(mail, envelope_header=False):
    """Returns a canonicalized email string you can use to send or store
    somewhere."""
    msg = to_message(mail).as_string(envelope_header)
    assert "From nobody" not in msg
    return msg


def from_string(data):
    """Takes a string, and tries to clean it up into a clean MailBase."""
    try:
        msg = email.message_from_string(data)
    except TypeError:
        msg = email.message_from_bytes(data)
    return from_message(msg)


def to_file(mail, fileobj):
    """Writes a canonicalized message to the given file."""
    fileobj.write(to_string(mail))


def from_file(fileobj):
    """Reads an email and cleans it up to make a MailBase."""
    try:
        msg = email.message_from_file(fileobj)
    except TypeError:
        fileobj.seek(0)
        msg = email.message_from_binary_file(fileobj)
    return from_message(msg)


def normalize_header(header):
    return string.capwords(header.lower(), '-')


def parse_parameter_header(message, header):
    params = message.get_params(header=header)
    if params:
        value = params.pop(0)[0]
        params_dict = dict(params)

        for key in CONTENT_ENCODING_REMOVED_PARAMS:
            if key in params_dict:
                del params_dict[key]

        return value, params_dict
    else:
        return None, {}


def properly_encode_header(value, encoder, not_email):
    """
    The only thing special (weird) about this function is that it tries
    to do a fast check to see if the header value has an email address in
    it.  Since random headers could have an email address, and email addresses
    have weird special formatting rules, we have to check for it.

    Normally this works fine, but in Librelist, we need to "obfuscate" email
    addresses by changing the '@' to '-AT-'.  This is where
    VALUE_IS_EMAIL_ADDRESS exists.  It's a simple lambda returning True/False
    to check if a header value has an email address.  If you need to make this
    check different, then change this.
    """
    try:
        value.encode("ascii")
        return value
    except UnicodeEncodeError:
        if not_email is False and VALUE_IS_EMAIL_ADDRESS(value):
            # this could have an email address, make sure we don't screw it up
            name, address = parseaddr(value)
            if six.PY2:
                # python 2 decodes to ascii, python 3 wants no decoding at all!
                name = name.encode("utf-8")
            return '"%s" <%s>' % (encoder.header_encode(name), address)

        if six.PY2:
            # python 2 decodes to ascii, python 3 wants no decoding at all!
            value = value.encode("utf-8")
        return "%s" % encoder.header_encode(value)


def header_to_mime_encoding(value, not_email=False):
    if not value:
        return ""

    encoder = Charset(DEFAULT_ENCODING)
    if isinstance(value, list):
        return "; ".join(properly_encode_header(v, encoder, not_email) for v in value)
    else:
        return properly_encode_header(value, encoder, not_email)


def header_from_mime_encoding(header):
    if header is None:
        return header
    elif isinstance(header, list):
        return [properly_decode_header(h) for h in header]
    elif isinstance(header, email.header.Header):
        return six.text_type(header)
    else:
        return properly_decode_header(header)


def guess_encoding_and_decode(original, data, errors=DEFAULT_ERROR_HANDLING):
    try:
        charset = chardet.detect(data)

        if not charset['encoding']:
            raise EncodingError("Header claimed %r charset, but detection found none. Decoding failed." % original)

        return data.decode(charset["encoding"], errors)
    except UnicodeError as exc:
        raise EncodingError("Header lied and claimed %r charset, guessing said "
                            "%r charset, neither worked so this is a bad email: "
                            "%s." % (original, charset, exc))


def attempt_decoding(charset, dec):
    """Attempts to decode bytes into unicode, calls guess_encoding_and_decode
    if the given charset is wrong."""
    try:
        if isinstance(dec, six.text_type):
            # it's already unicode so just return it
            return dec
        else:
            return dec.decode(charset)
    except (UnicodeError, LookupError):
        # looks like the charset lies, try to detect it
        return guess_encoding_and_decode(charset, dec)


def apply_charset_to_header(charset, encoding, data):
    """Given a charset and encoding, decode data into unicode, e.g.

        >>> print(apply_charset_to_header("utf-8", "Q", "=142ukasz"))
        łukasz

    ``encoding`` is case insensitive and must be one of B or Q
    """
    if encoding.upper() == 'B':
        dec = email.base64mime.decode(data.encode('ascii'))
    elif encoding.upper() == 'Q':
        if six.PY2:
            # python 2 decodes to ascii, python 3 wants no decoding at all!
            data = data.encode('ascii')
        dec = email.quoprimime.header_decode(data)
        if six.PY3:
            # and Python 3 gives us some bytes encoded as unicode chars, so encode them to bytes
            dec = bytes(dec, "raw-unicode-escape")
    else:
        raise EncodingError("Invalid header encoding %r should be 'Q' or 'B'." % encoding)

    return attempt_decoding(charset, dec)


def _match(data, pattern, pos):
    found = pattern.search(data, pos)
    if found:
        # there might be text that doesn't need decoding between the end of the
        # last match (pos) and the start of the new one (found.start())
        before_match = data[pos:found.start()]
        return before_match, found.groups(), found.end()
    else:
        before_match = data[pos:]
        return before_match, None, -1


def _tokenize(data, next_token):
    enc_data = None

    before_match, enc_header, next_token = _match(data, ENCODING_REGEX, next_token)

    if next_token != -1:
        enc_data, _, next_token = _match(data, ENCODING_END_REGEX, next_token)

    return before_match, enc_header, enc_data, next_token


def _scan(data):
    next_token = 0
    continued = False

    while next_token != -1:
        before_match, enc_header, enc_data, next_token = _tokenize(data, next_token)
        continued = next_token != -1 and INDENT_REGEX.match(data, next_token)

        yield before_match, enc_header, enc_data, continued


def _parse_charset_header(data):
    """Decodes header, yielding decoded and plain text sections separately

    For example:

        >>> data = '=?utf-8?q?=C5=81ukasz?= the =?utf-16?b?//492B/c?='
        >>> print(list(_parse_charset_header(data)))
        [u'\u0141ukasz', u' the ', u'\U0001f41f']
    """
    scanner = _scan(data)
    oddness = None

    try:
        while True:
            if not oddness:
                before_match, enc_header, enc_data, continued = six.next(scanner)
            else:
                before_match, enc_header, enc_data, continued = oddness
                oddness = None

            while continued:
                bm, eh, ed, continued = six.next(scanner)

                if not eh:
                    assert not ed, "Parsing error: %r" % data
                    oddness = (" " + bm.lstrip(), eh, ed, continued)
                elif eh[0] == enc_header[0] and eh[1] == enc_header[1]:
                    enc_data += ed
                else:
                    # odd case, it's continued but not from the same base64
                    # need to stack this for the next loop, and drop the \n\s+
                    oddness = ('', eh, ed, continued)
                    break

            if before_match:
                # bytes that weren't encoded to the "before_match", so we can decode it
                # directly to unicode
                yield attempt_decoding('ascii', before_match)

            if enc_header:
                yield apply_charset_to_header(enc_header[0], enc_header[1], enc_data)

    except StopIteration:
        pass


def properly_decode_header(header):
    """Decodes headers from their ASCII-safe representation"""
    return u"".join(_parse_charset_header(header))



"""
Confirmation handling API that helps you get the whole confirm/pending/verify
process correct.  It doesn't implement any handlers, but what it does do is
provide the logic for doing the following:

* Take an email, put it in a "pending" queue, and then send out a confirm
  email with a strong random id.
* Store the pending message ID and the random secret someplace for later
  verification.
* Verify an incoming email against the expected ID, and get back the
  original.

You then just work this into your project's state flow, write your own
templates, and possibly write your own storage.
"""
from __future__ import print_function, unicode_literals

from email.utils import parseaddr
import uuid

from salmon import queue, view


class ConfirmationStorage(object):
    """
    This is the basic confirmation storage.  For simple testing purposes
    you can just use the default hash db parameter.  If you do a deployment
    you can probably get away with a shelf hash instead.

    You can write your own version of this and use it.  The confirmation engine
    only cares that it gets something that supports all of these methods.
    """
    def __init__(self, db={}):
        """
        Change the db parameter to a shelf to get persistent storage.
        """
        self.confirmations = db

    def clear(self):
        """
        Used primarily in testing, this clears out all pending confirmations.
        """
        self.confirmations.clear()

    def key(self, target, from_address):
        """
        Used internally to construct a string key, if you write
        your own you don't need this.

        NOTE: To support proper equality and shelve storage, this encodes the
        key into ASCII.  Make a different subclass if you need Unicode and your
        storage supports it.
        """
        key = target + ':' + from_address

        return key.encode('ascii')

    def get(self, target, from_address):
        """
        Given a target and a from address, this returns a tuple of (expected_secret, pending_message_id).
        If it doesn't find that target+from_address, then it should return a (None, None) tuple.
        """
        return self.confirmations.get(self.key(target, from_address), (None, None))

    def delete(self, target, from_address):
        """
        Removes a target+from_address from the storage.
        """
        try:
            del self.confirmations[self.key(target, from_address)]
        except KeyError:
            pass

    def store(self, target, from_address, expected_secret, pending_message_id):
        """
        Given a target, from_address it will store the expected_secret and pending_message_id
        of later verification.  The target should be a string indicating what is being
        confirmed.  Like "subscribe", "post", etc.

        When implementing your own you should *never* allow more than one target+from_address
        combination.
        """
        self.confirmations[self.key(target, from_address)] = (expected_secret,
                                                              pending_message_id)


class ConfirmationEngine(object):
    """
    The confirmation engine is what does the work of sending a confirmation,
    and verifying that it was confirmed properly.  In order to use it you
    have to construct the ConfirmationEngine (usually in settings module) and
    you write your confirmation message templates for sending.

    The primary methods you use are ConfirmationEngine.send and ConfirmationEngine.verify.
    """
    def __init__(self, pending_queue, storage):
        """
        The pending_queue should be a string with the path to the salmon.queue.Queue
        that will store pending messages.  These messages are the originals the user
        sent when they tried to confirm.

        Storage should be something that is like ConfirmationStorage so that this
        can store things for later verification.
        """
        self.pending = queue.Queue(pending_queue)
        self.storage = storage

    def get_pending(self, pending_id):
        """
        Returns the pending message for the given ID.
        """
        return self.pending.get(pending_id)

    def push_pending(self, message):
        """
        Puts a pending message into the pending queue.
        """
        return self.pending.push(message)

    def delete_pending(self, pending_id):
        """
        Removes the pending message from the pending queue.
        """
        self.pending.remove(pending_id)

    def cancel(self, target, from_address, expect_secret):
        """
        Used to cancel a pending confirmation.
        """
        name, addr = parseaddr(from_address)

        secret, pending_id = self.storage.get(target, addr)

        if secret == expect_secret:
            self.storage.delete(target, addr)
            self.delete_pending(pending_id)

    def make_random_secret(self):
        """
        Generates a random uuid as the secret, in hex form.
        """
        return uuid.uuid4().hex

    def register(self, target, message):
        """
        Don't call this directly unless you know what you are doing.
        It does the job of registering the original message and the
        expected confirmation into the storage.
        """
        from_address = message.From

        pending_id = self.push_pending(message)
        secret = self.make_random_secret()
        self.storage.store(target, from_address, secret, pending_id)

        return "%s-confirm-%s" % (target, secret)

    def verify(self, target, from_address, expect_secret):
        """
        Given a target (i.e. "subscribe", "post", etc), a from_address
        of someone trying to confirm, and the secret they should use, this
        will try to verify their confirmation.  If the verify works then
        you'll get the original message back to do what you want with.

        If the verification fails then you are given None.

        The message is *not* deleted from the pending queue.  You can do
        that yourself with delete_pending.
        """
        assert expect_secret, "Must give an expected ID number."
        name, addr = parseaddr(from_address)

        secret, pending_id = self.storage.get(target, addr)

        if secret == expect_secret:
            self.storage.delete(target, addr)
            return self.get_pending(pending_id)
        else:
            return None

    def send(self, relay, target, message, template, vars):
        """
        This is the method you should use to send out confirmation messages.
        You give it the relay, a target (i.e. "subscribe"), the message they
        sent requesting the confirm, your confirmation template, and any
        vars that template needs.

        The result of calling this is that the template message gets sent through
        the relay, the original message is stored in the pending queue, and
        data is put into the storage for later calls to verify.
        """
        confirm_address = self.register(target, message)
        vars.update(locals())
        msg = view.respond(vars, template, To=message['from'],
                           From="%(confirm_address)s@%(host)s",
                           Subject="Confirmation required")

        msg['Reply-To'] = "%(confirm_address)s@%(host)s" % vars

        relay.deliver(msg)

    def clear(self):
        """
        Used in testing to make sure there's nothing in the pending
        queue or storage.
        """
        self.pending.clear()
        self.storage.clear()



# Copyright (C) 2008 Zed A. Shaw.  Licensed under the terms of the GPLv3.
from __future__ import print_function

from nose.tools import assert_equal, assert_raises
from salmon import encoding, mail

sample_message = """From: somedude@localhost
To: somedude@localhost

Test
"""


def test_mail_request():
    # try with a half-assed message
    msg = mail.MailRequest("localhost", "zedfrom@localhost",
                           "zedto@localhost", "Fake body.")
    assert_equal(msg['to'], "zedto@localhost")
    assert_equal(msg['from'], "zedfrom@localhost")

    msg = mail.MailRequest("localhost", "somedude@localhost", ["somedude@localhost"], sample_message)
    assert_equal(msg.original, sample_message)

    assert_equal(msg['From'], "somedude@localhost")

    assert("From" in msg)
    del msg["From"]
    assert("From" not in msg)

    msg["From"] = "nobody@localhost"
    assert("From" in msg)
    assert_equal(msg["From"], "nobody@localhost")
    msg["From"] = "somebody@localhost"
    assert_equal(msg["From"], "somebody@localhost")
    assert_equal(msg.keys(), ["To", "From"])
    assert_equal(msg.items(), [("To", "somedude@localhost"), ("From", "somebody@localhost")])

    # appending headers
    msg.base.append_header("To", "nobody@example.com")
    assert_equal(msg["To"], "somedude@localhost")
    assert_equal(msg.keys(), ["To", "From", "To"])
    assert_equal(msg.items(), [("To", "somedude@localhost"), ("From", "somebody@localhost"),
                 ("To", "nobody@example.com")])

    # validate that upper and lower case work for headers
    assert("FroM" in msg)
    assert("from" in msg)
    assert("From" in msg)
    assert_equal(msg['From'], msg['fRom'])
    assert_equal(msg['From'], msg['from'])
    assert_equal(msg['from'], msg['fRom'])

    # make sure repr runs
    print(repr(msg))

    return msg


def test_mail_response_plain_text():
    sample = mail.MailResponse(
        To="receiver@localhost",
        Subject="Test message",
        From="sender@localhost",
        Body="Test from test_mail_response_plain_text.",
    )

    return sample


def test_mail_response_html():
    sample = mail.MailResponse(
        To="receiver@localhost",
        Subject="Test message",
        From="sender@localhost",
        Html="<html><body><p>From test_mail_response_html</p></body></html>",
    )

    return sample


def test_mail_response_html_and_plain_text():
    sample = mail.MailResponse(
        To="receiver@localhost",
        Subject="Test message",
        From="sender@localhost",
        Html="<html><body><p>Hi there.</p></body></html>",
        Body="Test from test_mail_response_html_and_plain_text.",
    )

    return sample


def test_mail_response_attachments():
    sample = mail.MailResponse(
        To="receiver@localhost",
        Subject="Test message",
        From="sender@localhost",
        Body="Test from test_mail_response_attachments.",
    )
    readme_data = open("./README.rst").read()

    assert_raises(AssertionError, sample.attach, data=readme_data, disposition="inline")

    sample.attach(filename="./README.rst", content_type="text/plain", disposition="inline")
    assert_equal(len(sample.attachments), 1)
    assert sample.multipart

    msg = sample.to_message()
    assert_equal(len(msg.get_payload()), 2)

    sample.clear()
    assert_equal(len(sample.attachments), 0)
    assert not sample.multipart

    sample.attach(data=readme_data, filename="./README.rst", content_type="text/plain")

    msg = sample.to_message()
    assert_equal(len(msg.get_payload()), 2)
    sample.clear()

    sample.attach(data=readme_data, content_type="text/plain")
    msg = sample.to_message()
    assert_equal(len(msg.get_payload()), 2)

    return sample


def test_mail_request_attachments():
    sample = test_mail_response_attachments()
    data = str(sample)

    msg = mail.MailRequest("localhost", None, None, data)

    msg_parts = msg.all_parts()
    sample_parts = sample.all_parts()

    readme = open("./README.rst").read()

    assert msg_parts[0].body == sample_parts[0].body
    assert readme == msg_parts[1].body
    assert msg.body() == sample_parts[0].body

    # test that we get at least one message for messages without attachments
    sample = test_mail_response_plain_text()
    msg = mail.MailRequest("localhost", None, None, str(sample))
    msg_parts = msg.all_parts()
    assert_equal(len(msg_parts), 0)
    assert msg.body()


def test_mail_response_mailing_list_headers():
    list_addr = "test.users@localhost"

    msg = mail.MailResponse(From='somedude@localhost', To=list_addr, Subject='subject', Body="Mailing list reply.")

    print(repr(msg))

    msg["Sender"] = list_addr
    msg["Reply-To"] = list_addr
    msg["List-Id"] = list_addr
    msg["Return-Path"] = list_addr
    msg["In-Reply-To"] = 'Message-Id-1231123'
    msg["References"] = 'Message-Id-838845854'
    msg["Precedence"] = 'list'

    data = str(msg)

    req = mail.MailRequest('localhost', 'somedude@localhost', list_addr, data)

    headers = ['Sender', 'Reply-To', 'List-Id', 'Return-Path', 'In-Reply-To', 'References', 'Precedence']
    for header in headers:
        assert msg[header] == req[header], "%s: %r != %r" % (header, msg[header], req[header])

    # try a delete
    del msg['Precedence']


def test_mail_response_headers():
    msg = test_mail_response_plain_text()
    # validate that upper and lower case work for headers
    assert("FroM" in msg)
    assert("from" in msg)
    assert("From" in msg)
    assert_equal(msg['From'], msg['fRom'])
    assert_equal(msg['From'], msg['from'])
    assert_equal(msg['from'], msg['fRom'])

    assert_equal(msg.keys(), [i[0] for i in msg.items()])


def test_walk():
    bm = mail.MailRequest(None, None, None, open("tests/bounce.msg").read())
    parts = [x for x in bm.walk()]

    assert parts
    assert_equal(len(parts), 6)


def test_copy_parts():
    bm = mail.MailRequest(None, None, None, open("tests/bounce.msg").read())

    resp = mail.MailResponse(To=bm['to'], From=bm['from'], Subject=bm['subject'])

    resp.attach_all_parts(bm)

    resp = resp.to_message()
    bm = bm.to_message()

    assert_equal(len([x for x in bm.walk()]), len([x for x in resp.walk()]))


def test_craft_from_sample():
    list_name = "test.list"
    user_full_address = "tester@localhost"

    sample = mail.MailResponse(
        To=list_name + "@localhost",
        From=user_full_address,
        Subject="Test message with attachments.",
        Body="The body as one attachment.",
    )
    sample.update({"Test": "update"})

    sample.attach(filename="tests/salmon_tests/message_tests.py", content_type="text/plain", disposition="attachment")

    inmsg = mail.MailRequest("fakepeer", None, None, str(sample))
    assert "Test" in sample.keys()

    for part in inmsg.to_message().walk():
        assert part.get_payload(), "inmsg busted."

    outmsg = mail.MailResponse(To=inmsg['from'], From=inmsg['to'], Subject=inmsg['subject'])

    outmsg.attach_all_parts(inmsg)

    result = outmsg.to_message()

    for part in result.walk():
        assert part.get_payload(), "outmsg parts don't have payload."


def test_to_from_works():
    msg = mail.MailRequest("fakepeer", "from@localhost", [u"<to1@localhost>", u"to2@localhost"], "")
    assert '<' not in msg.To, msg.To

    msg = mail.MailRequest("fakepeer", "from@localhost", [u"to1@localhost", u"to2@localhost"], "")
    assert '<' not in msg.To, msg.To

    msg = mail.MailRequest("fakepeer", "from@localhost", [u"to1@localhost", u"<to2@localhost>"], "")
    assert '<' not in msg.To, msg.To

    msg = mail.MailRequest("fakepeer", "from@localhost", [u"to1@localhost"], "")
    assert '<' not in msg.To, msg.To

    msg = mail.MailRequest("fakepeer", "from@localhost", [u"<to1@localhost>"], "")
    assert '<' not in msg.To, msg.To


def test_decode_header_randomness():
    assert_equal(mail._decode_header_randomness(None), set())

    # with strings
    assert_equal(mail._decode_header_randomness(["z@localhost", '"Z A" <z@localhost>']),
                 set(["z@localhost", "z@localhost"]))
    assert_equal(mail._decode_header_randomness("z@localhost"), set(["z@localhost"]))

    # with bytes
    assert_equal(mail._decode_header_randomness([b"z@localhost", b'"Z A" <z@localhost>']),
                 set(["z@localhost", "z@localhost"]))
    assert_equal(mail._decode_header_randomness(b"z@localhost"), set(["z@localhost"]))

    # with literal nonsense
    assert_raises(encoding.EncodingError, mail._decode_header_randomness, 1)
    assert_raises(encoding.EncodingError, mail._decode_header_randomness, [1, "m@localhost"])



from salmon import queue
from salmon.routing import nolocking, route, stateless
from salmon.utils import settings


@route("(to)@(host)", to=".+", host="example.com")
@stateless
@nolocking
def START(message, to=None, host=None):
    inbox = queue.Queue(settings.QUEUE_PATH)
    inbox.push(message)



from nose.tools import assert_equal, with_setup
from salmon import mail
from salmon.routing import Router

from .handlers import bounce_filtered_mod
from .setup_env import setup_salmon_dirs, teardown_salmon_dirs


def test_bounce_analyzer_on_bounce():
    bm = mail.MailRequest(None, None, None, open("tests/bounce.msg").read())
    assert bm.is_bounce()
    assert bm.bounce
    assert_equal(bm.bounce.score, 1.0)
    assert bm.bounce.probable()
    assert_equal(bm.bounce.primary_status, (5, u'Permanent Failure'))
    assert_equal(bm.bounce.secondary_status, (1, u'Addressing Status'))
    assert_equal(bm.bounce.combined_status, (11, u'Bad destination mailbox address'))

    assert bm.bounce.is_hard()
    assert_equal(bm.bounce.is_hard(), not bm.bounce.is_soft())

    assert_equal(bm.bounce.remote_mta, u'gmail-smtp-in.l.google.com')
    assert_equal(bm.bounce.reporting_mta, u'mail.zedshaw.com')
    assert_equal(bm.bounce.final_recipient,
                 u'asdfasdfasdfasdfasdfasdfewrqertrtyrthsfgdfgadfqeadvxzvz@gmail.com')
    assert_equal(bm.bounce.diagnostic_codes[0], u'550-5.1.1')
    assert_equal(bm.bounce.action, 'failed')
    assert 'Content-Description-Parts' in bm.bounce.headers

    assert bm.bounce.error_for_humans()


def test_bounce_analyzer_on_regular():
    bm = mail.MailRequest(None, None, None, open("tests/signed.msg").read())
    assert not bm.is_bounce()
    assert bm.bounce
    assert_equal(bm.bounce.score, 0.0)
    assert not bm.bounce.probable()
    assert_equal(bm.bounce.primary_status, (None, None))
    assert_equal(bm.bounce.secondary_status, (None, None))
    assert_equal(bm.bounce.combined_status, (None, None))

    assert not bm.bounce.is_hard()
    assert not bm.bounce.is_soft()

    assert_equal(bm.bounce.remote_mta, None)
    assert_equal(bm.bounce.reporting_mta, None)
    assert_equal(bm.bounce.final_recipient, None)
    assert_equal(bm.bounce.diagnostic_codes, [None, None])
    assert_equal(bm.bounce.action, None)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_bounce_to_decorator():
    msg = mail.MailRequest(None, None, None, open("tests/bounce.msg").read())

    Router.deliver(msg)
    assert Router.in_state(bounce_filtered_mod.START, msg)
    assert bounce_filtered_mod.HARD_RAN, "Hard bounce state didn't actually run: %r" % msg.route_to

    msg.bounce.primary_status = (4, u'Persistent Transient Failure')
    Router.clear_states()
    Router.deliver(msg)
    assert Router.in_state(bounce_filtered_mod.START, msg)
    assert bounce_filtered_mod.SOFT_RAN, "Soft bounce didn't actually run."

    msg = mail.MailRequest(None, None, None, open("tests/signed.msg").read())
    Router.clear_states()
    Router.deliver(msg)
    assert Router.in_state(bounce_filtered_mod.END, msg), "Regular messages aren't delivering."


def test_bounce_getting_original():
    msg = mail.MailRequest(None, None, None, open("tests/bounce.msg").read())
    msg.is_bounce()

    assert msg.bounce.notification
    assert msg.bounce.notification.body

    assert msg.bounce.report

    for part in msg.bounce.report:
        assert [(k, part[k]) for k in part]
        # these are usually empty, but might not be.  they are in our test
        assert not part.body

    assert msg.bounce.original
    assert_equal(msg.bounce.original['to'], msg.bounce.final_recipient)
    assert msg.bounce.original.body


def test_bounce_no_headers_error_message():
    msg = mail.MailRequest(None, None, None, "Nothing")
    msg.is_bounce()
    assert_equal(msg.bounce.error_for_humans(), 'No status codes found in bounce message.')



from __future__ import print_function, unicode_literals

import argparse
import email
import glob
import mailbox
import os
import shutil
import signal
import sys

from salmon import encoding, mail, queue, routing, server, utils
import salmon

COMMANDS = (
    ("start", "starts a server"),
    ("stop", "stops a server"),
    ("status", "displays status of server"),
    ("gen", "generate new project"),
    ("log", "start log server"),
    ("queue", "manipulate a Queue"),
    ("blast", "blast emails at a server"),
    ("cleanse", "cleanse your emails"),
    ("routes", "display routes"),
    ("send", "send a new email"),
    ("sendmail", "send an email from stdin"),
)

DEFAULT_PID_FILE = "./run/stmp.pid"

version_info = """
Salmon-Version:  %s
Version-File:  %s
""" % (salmon.__version__, salmon.__file__)

copyright_notice = """
Salmon is Copyright (C) Matt Molyneaux 2014-2015.  Licensed GPLv3.
Forked from Lamon, Copyright (C) Zed A. Shaw 2008-2009.  Licensed GPLv3.
If you didn't get a copy of the LICENSE go to:

    https://github.com/moggers87/salmon/LICENSE

Have fun.
"""

uid_desc = """
If you specify a uid/gid then this means you want to first change to
root, set everything up, and then drop to that UID/GID combination.
This is typically so you can bind to port 25 and then become "safe"
to continue operating as a non-root user. If you give one or the other,
this it will just change to that uid or gid without doing the priv drop
operation.
"""


def main():
    """Salmon script entry point"""
    args = _parser.parse_args()

    # get function reference from args and then remove it
    cmd = args.func
    del args.func

    # pass all other attrs to function as kwargs
    cmd(**vars(args))


def log_command(parser):
    """
    Runs a logging only server on the given hosts and port.  It logs
    each message it receives and also stores it to the run/queue
    so that you can make sure it was received in testing.
    """
    def command(port, host, pid, chdir, chroot=None, uid=False, gid=False, umask=False,
                force=False, debug=False, daemon=True):
        utils.start_server(pid, force, chroot, chdir, uid, gid, umask,
                           lambda: utils.make_fake_settings(host, port), debug, daemon)

    parser.set_defaults(func=command)

    parser.add_argument("--port", default=8825, type=int, help="port to listen on")
    parser.add_argument("--host", default="127.0.0.1", help="address to listen on")
    parser.add_argument("--chroot", default=argparse.SUPPRESS, help="path to chroot")
    parser.add_argument("--chdir", default=".", help="change to this directory when daemonising")
    parser.add_argument("--umask", type=int, default=argparse.SUPPRESS, help="set umask on server")
    parser.add_argument("--pid", default="./run/log.pid", help="path to pid file")
    parser.add_argument("-f", "--force", action="store_true", help="force server to run, ignoring pid file")
    parser.add_argument("--debug", action="store_true", default=argparse.SUPPRESS, help="debug mode")

    uid_group = parser.add_argument_group(title="UID/GID options", description=uid_desc)
    uid_group.add_argument("--uid", type=int, default=argparse.SUPPRESS, help="run with this user id")
    uid_group.add_argument("--gid", type=int, default=argparse.SUPPRESS, help="run with this group id")

    daemon_group = parser.add_mutually_exclusive_group()
    daemon_group.add_argument("--no-daemon", default=argparse.SUPPRESS, dest="daemon", action="store_false",
                              help="start server in foreground")
    daemon_group.add_argument("--daemon", default=argparse.SUPPRESS, dest="daemon", action="store_true",
                              help="start server as daemon (default)")


def send_command(parser):
    """
    Sends an email to someone as a test message.
    See the sendmail command for a sendmail replacement.
    """
    def command(port, host, username=None, password=None, ssl=None, starttls=None, lmtp=None,
                sender=None, to=None, subject=None, body=None, attach=None):
        message = mail.MailResponse(From=sender, To=to, Subject=subject, Body=body)
        if attach:
            message.attach(attach)

        relay = server.Relay(host, port=port, username=username, password=password, ssl=ssl,
                             starttls=starttls, lmtp=lmtp, debug=False)
        relay.deliver(message)

    parser.set_defaults(func=command)

    parser.add_argument("--port", default=8825, type=int, help="port to listen on")
    parser.add_argument("--host", default="127.0.0.1", help="address to listen on")
    parser.add_argument("--username", default=argparse.SUPPRESS, help="SMTP username")
    parser.add_argument("--password", default=argparse.SUPPRESS, help="SMTP password")
    parser.add_argument("--sender", metavar="EMAIL", default=argparse.SUPPRESS)
    parser.add_argument("--to", metavar="EMAIL", default=argparse.SUPPRESS)
    parser.add_argument("--subject", default=argparse.SUPPRESS)
    parser.add_argument("--body", default=argparse.SUPPRESS)
    parser.add_argument("--attach", default=argparse.SUPPRESS)
    parser.add_argument("--lmtp", action="store_true", default=argparse.SUPPRESS)

    tls_group = parser.add_mutually_exclusive_group()
    tls_group.add_argument("--ssl", action="store_true", default=argparse.SUPPRESS)
    tls_group.add_argument("--starttls", action="store_true", default=argparse.SUPPRESS)


def sendmail_command(parser):
    """
    Used as a testing sendmail replacement for use in programs
    like mutt as an MTA.  It reads the email to send on the stdin
    and then delivers it based on the port and host settings.
    """
    def command(port, host, recipients, debug=False, lmtp=False):
        relay = server.Relay(host, port=port, debug=debug, lmtp=lmtp)
        data = sys.stdin.read()
        msg = mail.MailRequest(None, recipients, None, data)
        relay.deliver(msg)

    parser.set_defaults(func=command)

    parser.add_argument("--port", default=8825, type=int, help="Port to listen on")
    parser.add_argument("--host", default="127.0.0.1", help="Address to listen on")
    parser.add_argument("--lmtp", action="store_true", default=argparse.SUPPRESS, help="Use LMTP rather than SMTP")
    parser.add_argument("--debug", action="store_true", default=argparse.SUPPRESS, help="Debug mode")
    parser.add_argument("recipients", action="append")


def start_command(parser):
    """
    Runs a salmon server out of the current directory
    """
    def command(pid, force, chdir, boot, chroot=False, uid=False, gid=False, umask=False, debug=False, daemon=True):
        utils.start_server(pid, force, chroot, chdir, uid, gid, umask,
                           lambda: utils.import_settings(True, boot_module=boot), debug, daemon)

    parser.set_defaults(func=command)

    parser.add_argument("--boot", default="config.boot", help="module with server definition")
    parser.add_argument("--chroot", default=argparse.SUPPRESS, help="path to chroot")
    parser.add_argument("--chdir", default=".", help="change to this directory when daemonising")
    parser.add_argument("--umask", type=int, default=argparse.SUPPRESS, help="set umask on server")
    parser.add_argument("--pid", default=DEFAULT_PID_FILE, help="path to pid file")
    parser.add_argument("-f", "--force", action="store_true", help="force server to run, ignoring pid file")
    parser.add_argument("--debug", action="store_true", default=argparse.SUPPRESS, help="debug mode")

    uid_group = parser.add_argument_group(title="UID/GID options", description=uid_desc)
    uid_group.add_argument("--uid", type=int, default=argparse.SUPPRESS, help="run with this user id")
    uid_group.add_argument("--gid", type=int, default=argparse.SUPPRESS, help="run with this group id")

    daemon_group = parser.add_mutually_exclusive_group()
    daemon_group.add_argument("--no-daemon", default=argparse.SUPPRESS, dest="daemon", action="store_false",
                              help="start server in foreground")
    daemon_group.add_argument("--daemon", default=argparse.SUPPRESS, dest="daemon", action="store_true",
                              help="start server as daemon (default)")


def stop_command(parser):
    """
    Stops a running salmon server
    """
    def command(pid, kill=False, all=False):
        pid_files = []

        if all:
            pid_files = glob.glob(all + "/*.pid")
        else:
            pid_files = [pid]

            if not os.path.exists(pid):
                print("PID file %s doesn't exist, maybe Salmon isn't running?" % pid)
                sys.exit(1)
                return  # for unit tests mocking sys.exit

        print("Stopping processes with the following PID files: %s" % pid_files)

        for pid_f in pid_files:
            pid = open(pid_f).readline()

            print("Attempting to stop salmon at pid %d" % int(pid))

            try:
                if kill:
                    os.kill(int(pid), signal.SIGKILL)
                else:
                    os.kill(int(pid), signal.SIGHUP)

                os.unlink(pid_f)
            except OSError as exc:
                print("ERROR stopping Salmon on PID %d: %s" % (int(pid), exc))

    parser.set_defaults(func=command)

    parser.add_argument("--pid", default=DEFAULT_PID_FILE, help="path to pid file")
    parser.add_argument("-f", "--force", dest="kill", default=DEFAULT_PID_FILE, action="store_true",
                        help="force stop server")
    parser.add_argument("--all", default=argparse.SUPPRESS,
                        help="stops all servers with .pid files in the specified directory")


def status_command(parser):
    """
    Prints out status information about salmon useful for finding out if it's
    running and where.
    """
    def command(pid):
        if os.path.exists(pid):
            pid = open(pid).readline()
            print("Salmon running with PID %d" % int(pid))
        else:
            print("Salmon not running.")

    parser.set_defaults(func=command)

    parser.add_argument("--pid", default=DEFAULT_PID_FILE, help="path to pid file")


def queue_command(parser):
    """
    Lets you do most of the operations available to a queue.
    """
    def command(name, pop=False, get=False, keys=False, remove=False, count=False, clear=False):
        print("Using queue: %r" % name)

        inq = queue.Queue(name)

        if pop:
            key, msg = inq.pop()
            if key:
                print("KEY: %s" % key)
                print(msg)
        elif get:
            print(inq.get(get))
        elif remove:
            inq.remove(remove)
        elif count:
            print("Queue %s contains %d messages" % (name, inq.count()))
        elif clear:
            inq.clear()
        elif keys:
            print("\n".join(inq.keys()))

    parser.set_defaults(func=command)

    command_group = parser.add_mutually_exclusive_group(required=True)
    command_group.add_argument("--pop", action="store_true", default=argparse.SUPPRESS,
                               help="pop a message from queue")
    command_group.add_argument("--get", metavar="KEY", default=argparse.SUPPRESS, help="get key from queue")
    command_group.add_argument("--remove", metavar="KEY", default=argparse.SUPPRESS,
                               help="remove chosen key from queue")
    command_group.add_argument("--count", action="store_true", default=argparse.SUPPRESS,
                               help="count messages in queue")
    command_group.add_argument("--clear", action="store_true", default=argparse.SUPPRESS, help="clear queue")
    command_group.add_argument("--keys", action="store_true", default=argparse.SUPPRESS, help="print queue keys")

    parser.add_argument("name", nargs="?", default="./run/queue", help="path of queue", metavar="queue")


def routes_command(parser):
    """
    Prints out valuable information about an application's routing configuration
    after everything is loaded and ready to go.  Helps debug problems with
    messages not getting to your handlers.  Path has the search paths you want
    separated by a ':' character, and it's added to the sys.path.
    """
    def command(modules, path=os.getcwd(), test=""):
        sys.path += path.split(':')
        test_case_matches = []

        for module in modules:
            __import__(module, globals(), locals())

        print("Routing ORDER: ", routing.Router.ORDER)
        print("Routing TABLE: \n---")
        for format in routing.Router.REGISTERED:
            print("%r: " % format, end="")
            regex, functions = routing.Router.REGISTERED[format]
            for func in functions:
                print("%s.%s " % (func.__module__, func.__name__), end="")
                match = regex.match(test)
                if test and match:
                    test_case_matches.append((format, func, match))

            print("\n---")

        if test_case_matches:
            print("\nTEST address %r matches:" % test)
            for format, func, match in test_case_matches:
                print("  %r %s.%s" % (format, func.__module__, func.__name__))
                print("  -  %r" % (match.groupdict()))
        elif test:
            print("\nTEST address %r didn't match anything." % test)

    parser.set_defaults(func=command)

    parser.add_argument("--path", default=argparse.SUPPRESS, help="search path for modules")
    parser.add_argument("modules", metavar="module", nargs="*", default=["config.testing"],
                        help="config modules to process")
    parser.add_argument("--test", metavar="EMAIL", default=argparse.SUPPRESS, help="test address")


def gen_command(parser):
    """
    Generates various useful things for you to get you started.
    """
    def command(project, force=False):
        template = os.path.join(salmon.__path__[0], "data", "prototype")

        if os.path.exists(project) and not force:
            print("Project %s exists, delete it first." % project)
            sys.exit(1)
            return
        elif force:
            shutil.rmtree(project, ignore_errors=True)

        shutil.copytree(template, project)

    parser.set_defaults(func=command)

    parser.add_argument("project", help="project name")
    parser.add_argument("-f", "--force", action="store_true", default=argparse.SUPPRESS,
                        help="overwrite existing directories")


def cleanse_command(parser):
    """
    Uses Salmon mail cleansing and canonicalization system to take an
    input Maildir (or mbox) and replicate the email over into another
    Maildir.  It's used mostly for testing and cleaning.
    """
    def command(input, output):
        error_count = 0

        try:
            inbox = mailbox.mbox(input)
        except IOError:
            inbox = mailbox.Maildir(input, factory=None)

        outbox = mailbox.Maildir(output)

        for msg in inbox:
            try:
                mail = encoding.from_message(msg)
                outbox.add(encoding.to_string(mail))
            except encoding.EncodingError as exc:
                print("ERROR: %s" % exc)
                error_count += 1

        outbox.close()
        inbox.close()

        print("TOTAL ERRORS: %s" % error_count)

    parser.set_defaults(func=command)

    parser.add_argument("input", help="input Maildir or mbox")
    parser.add_argument("output", help="output Maildir")


def blast_command(parser):
    """
    Given a Maildir, this command will go through each email
    and blast it at your server.  It does nothing to the message, so
    it will be real messages hitting your server, not cleansed ones.
    """
    def command(input, host, port, lmtp=None, debug=False):
        try:
            inbox = mailbox.mbox(input)
        except IOError:
            inbox = mailbox.Maildir(input, factory=None)

        relay = server.Relay(host, port=port, lmtp=lmtp, debug=debug)

        for key in inbox.keys():
            msgfile = inbox.get_file(key)
            msg = email.message_from_file(msgfile)
            relay.deliver(msg)

    parser.set_defaults(func=command)

    parser.add_argument("input", help="input Maildir or mbox")
    parser.add_argument("--port", default=8823, type=int, help="port to listen on")
    parser.add_argument("--host", default="127.0.0.1", help="address to listen on")
    parser.add_argument("--lmtp", action="store_true", default=argparse.SUPPRESS)
    parser.add_argument("--debug", action="store_true", default=argparse.SUPPRESS, help="debug mode")


# Bring it all together

_parser = argparse.ArgumentParser(description="Python mail server", epilog=copyright_notice,
                                  formatter_class=argparse.RawDescriptionHelpFormatter)

_parser.add_argument("-v", "--version", action="version", version=version_info)
_subparsers = _parser.add_subparsers(metavar="<command>")
_subparsers.required = True

for cmd, help_txt in COMMANDS:
    function = globals()["{0}_command".format(cmd)]
    cmd_parser = _subparsers.add_parser(cmd, description=function.__doc__, help=help_txt,
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    function(cmd_parser)



"""
Mostly utility functions Salmon uses internally that don't
really belong anywhere else in the modules.  This module
is kind of a dumping ground, so if you find something that
can be improved feel free to work up a patch.
"""
from __future__ import print_function, unicode_literals

import imp
import importlib
import logging
import os
import sys

from lockfile import pidlockfile
import daemon

from salmon import routing, server

settings = None


def import_settings(boot_also, boot_module="config.boot"):
    """Returns the current settings module, there is no harm in calling it
    multiple times

    The location of the settings module can be control via
    ``SALMON_SETTINGS_MODULE``"""
    global settings

    if settings is None:
        settings_module = os.getenv("SALMON_SETTINGS_MODULE", "config.settings")
        settings = importlib.import_module(settings_module)

    if boot_also:
        importlib.import_module(boot_module)

    return settings


def daemonize(pid, chdir, chroot, umask, files_preserve=None, do_open=True):
    """
    Uses python-daemonize to do all the junk needed to make a
    server a server.  It supports all the features daemonize
    has, except that chroot probably won't work at all without
    some serious configuration on the system.
    """
    context = daemon.DaemonContext()
    context.pidfile = pidlockfile.PIDLockFile(pid)
    context.stdout = open(os.path.join(chdir, "logs/salmon.out"), "a+")
    context.stderr = open(os.path.join(chdir, "logs/salmon.err"), "a+")
    context.files_preserve = files_preserve or []
    context.working_directory = os.path.expanduser(chdir)

    if chroot:
        context.chroot_directory = os.path.expanduser(chroot)
    if umask is not False:
        context.umask = umask

    if do_open:
        context.open()

    return context


def drop_priv(uid, gid):
    """
    Changes the uid/gid to the two given, you should give utils.daemonize
    0,0 for the uid,gid so that it becomes root, which will allow you to then
    do this.
    """
    logging.debug("Dropping to uid=%d, gid=%d", uid, gid)
    daemon.daemon.change_process_owner(uid, gid)
    logging.debug("Now running as uid=%d, gid=%d", os.getgid(), os.getuid())


def make_fake_settings(host, port):
    """
    When running as a logging server we need a fake settings module to work with
    since the logging server can be run in any directory, so there may not be
    a settings module to import.
    """
    global settings

    if settings is None:
        logging.basicConfig(filename="logs/logger.log", level=logging.DEBUG)
        routing.Router.load(['salmon.handlers.log', 'salmon.handlers.queue'])
        settings = imp.new_module('settings')
        settings.receiver = server.SMTPReceiver(host, port)
        settings.relay = None
        logging.info("Logging mode enabled, will not send email to anyone, just log.")

    return settings


def check_for_pid(pid, force):
    """Checks if a pid file is there, and if it is sys.exit.  If force given
    then it will remove the file and not exit if it's there."""
    if os.path.exists(pid):
        if not force:
            print("PID file %s exists, so assuming Salmon is running.  Give --force to force it to start." % pid)
            sys.exit(1)
        else:
            os.unlink(pid)


def start_server(pid, force, chroot, chdir, uid, gid, umask, settings_loader, debug, daemon_proc):
    """
    Starts the server by doing a daemonize and then dropping priv
    accordingly.  It will only drop to the uid/gid given if both are given.
    """
    check_for_pid(pid, force)

    if not debug and daemon_proc:
        daemonize(pid, chdir, chroot, umask, files_preserve=[])

    sys.path.append(os.getcwd())

    settings = settings_loader()

    if uid and gid:
        drop_priv(uid, gid)
    elif uid or gid:
        logging.warning("You probably meant to give a uid and gid, but you gave: uid=%r, gid=%r. "
                        "Will not change to any user.", uid, gid)

    settings.receiver.start()

    if debug:
        print("Salmon started in debug mode. ctrl-c to quit...")
        import time
        try:
            while True:
                time.sleep(100000)
        except KeyboardInterrupt:
            # hard quit, since receiver starts a new thread. dirty but works
            os._exit(1)



"""
Salmon comes with a few useful handlers that you can add to your salmon.routing
configuration.  Most of them are implemented as stateless handlers, but some
will require you to read the documentation and configure before you can use
them.
"""



"""
The majority of the server related things Salmon needs to run, like receivers,
relays, and queue processors.
"""
from __future__ import print_function, unicode_literals

import asyncore
import logging
import smtpd
import smtplib
import threading
import time
import traceback

from dns import resolver
import lmtpd
import six

from salmon import __version__, mail, queue, routing
from salmon.bounce import COMBINED_STATUS_CODES, PRIMARY_STATUS_CODES, SECONDARY_STATUS_CODES

lmtpd.__version__ = "Salmon Mail router LMTPD, version %s" % __version__
smtpd.__version__ = "Salmon Mail router SMTPD, version %s" % __version__

if six.PY2:
    # Python 2 commits many crimes against byte encoding, so we've got to manually convert to ASCII
    smtpd.__version__ = smtpd.__version__.encode()


def undeliverable_message(raw_message, failure_type):
    """
    Used universally in this file to shove totally screwed messages
    into the routing.Router.UNDELIVERABLE_QUEUE (if it's set).
    """
    if routing.Router.UNDELIVERABLE_QUEUE:
        key = routing.Router.UNDELIVERABLE_QUEUE.push(raw_message)

        logging.error("Failed to deliver message because of %r, put it in "
                      "undeliverable queue with key %r", failure_type, key)


class SMTPError(Exception):
    """
    You can raise this error when you want to abort with a SMTP error code to
    the client.  This is really only relevant when you're using the
    SMTPReceiver and the client understands the error.

    If you give a message than it'll use that, but it'll also produce a
    consistent error message based on your code.  It uses the errors in
    salmon.bounce to produce them.
    """
    def __init__(self, code, message=None):
        self.code = code
        self.message = message or self.error_for_code(code)

        Exception.__init__(self, "%d %s" % (self.code, self.message))

    def error_for_code(self, code):
        primary, secondary, tertiary = str(code)

        primary = PRIMARY_STATUS_CODES.get(primary, "")
        secondary = SECONDARY_STATUS_CODES.get(secondary, "")
        combined = COMBINED_STATUS_CODES.get(primary + secondary, "")

        return " ".join([primary, secondary, combined]).strip()


class Relay(object):
    """
    Used to talk to your "relay server" or smart host, this is probably the most
    important class in the handlers next to the salmon.routing.Router.
    It supports a few simple operations for sending mail, replying, and can
    log the protocol it uses to stderr if you set debug=1 on __init__.
    """
    def __init__(self, host='127.0.0.1', port=25, username=None, password=None,
                 ssl=False, starttls=False, debug=0, lmtp=False):
        """
        The hostname and port we're connecting to, and the debug level (default to 0).
        Optional username and password for smtp authentication.
        If ssl is True smtplib.SMTP_SSL will be used.
        If starttls is True (and ssl False), smtp connection will be put in TLS mode.
        If lmtp is true, then smtplib.LMTP will be used. Mutually exclusive with ssl.
        """
        self.hostname = host
        self.port = port
        self.debug = debug
        self.username = username
        self.password = password
        self.ssl = ssl
        self.starttls = starttls
        self.lmtp = lmtp

        assert not (ssl and lmtp), "LMTP over SSL not supported. Use STARTTLS instead."
        assert not (ssl and starttls), "SSL and STARTTLS make no sense together"

    def configure_relay(self, hostname):
        if self.ssl:
            relay_host = smtplib.SMTP_SSL(hostname, self.port)
        elif self.lmtp:
            relay_host = smtplib.LMTP(hostname, self.port)
        else:
            relay_host = smtplib.SMTP(hostname, self.port)

        relay_host.set_debuglevel(self.debug)

        if self.starttls:
            relay_host.starttls()
        if self.username and self.password:
            relay_host.login(self.username, self.password)

        assert relay_host, 'Code error, file a bug.'
        return relay_host

    def deliver(self, message, To=None, From=None):
        """
        Takes a fully formed email message and delivers it to the
        configured relay server.

        You can pass in an alternate To and From, which will be used in the
        SMTP/LMTP send lines rather than what's in the message.
        """
        # Check in multiple places for To and From.
        # Ordered in preference.
        recipient = To or getattr(message, 'To', None) or message['To']
        sender = From or getattr(message, 'From', None) or message['From']

        hostname = self.hostname or self.resolve_relay_host(recipient)

        relay_host = self.configure_relay(hostname)
        relay_host.sendmail(sender, recipient, str(message))
        relay_host.quit()

    def resolve_relay_host(self, To):
        target_host = To.split("@")[1]

        try:
            mx_host = str(resolver.query(target_host, "mx")[0].exchange)
        except resolver.NoAnswer:
            logging.debug("Domain %r does not have an MX record, using %r instead.", target_host, target_host)
            return target_host

        logging.debug("Delivering to MX record %r for target %r", mx_host, target_host)
        return mx_host

    def __repr__(self):
        """Used in logging and debugging to indicate where this relay goes."""
        return "<Relay to (%s:%d)>" % (self.hostname, self.port)

    def reply(self, original, From, Subject, Body):
        """Calls self.send but with the from and to of the original message reversed."""
        self.send(original.From, From=From, Subject=Subject, Body=Body)

    def send(self, To, From, Subject, Body):
        """
        Does what it says, sends an email.  If you need something more complex
        then look at salmon.mail.MailResponse.
        """
        msg = mail.MailResponse(To=To, From=From, Subject=Subject, Body=Body)
        self.deliver(msg)


class SMTPChannel(smtpd.SMTPChannel):
    """Replaces the standard SMTPChannel with one that rejects more than one recipient"""

    def smtp_RCPT(self, arg):
        if self.__rcpttos:
            # We can't properly handle multiple RCPT TOs in SMTPReceiver
            #
            # SMTP can only return one reply at the end of DATA, making it an
            # all or nothing reply. As we can't roll back a previously
            # successful delivery and the delivery happens without there being
            # a queue, we can end up in a state where one recipient has
            # received their mail and another has not (due to a 550 response
            # raised by the handler). At that point there's no reasonable
            # response to give the client - we haven't delivered everything,
            # but we haven't delivered *nothing* either.
            #
            # So we bug out early and hope for the best. At worst mail will
            # bounce, but nothing will be lost.
            #
            # Of course, if smtpd.SMTPServer or SMTPReceiver implemented a
            # queue and bounces like you're meant too...
            logging.warning("Client attempted to deliver mail with multiple RCPT TOs. This is not supported.")
            self.push("451 Will not accept multiple recipients in one transaction")
        else:
            smtpd.SMTPChannel.smtp_RCPT(self, arg)


class SMTPReceiver(smtpd.SMTPServer):
    """Receives emails and hands it to the Router for further processing."""

    def __init__(self, host='127.0.0.1', port=8825):
        """
        Initializes to bind on the given port and host/IP address.  Typically
        in deployment you'd give 0.0.0.0 for "all internet devices" but consult
        your operating system.

        This uses smtpd.SMTPServer in the __init__, which means that you have to
        call this far after you use python-daemonize or else daemonize will
        close the socket.
        """
        self.host = host
        self.port = port
        smtpd.SMTPServer.__init__(self, (self.host, self.port), None)

    def start(self):
        """
        Kicks everything into gear and starts listening on the port.  This
        fires off threads and waits until they are done.
        """
        logging.info("SMTPReceiver started on %s:%d.", self.host, self.port)
        self.poller = threading.Thread(target=asyncore.loop, kwargs={'timeout': 0.1, 'use_poll': True})
        self.poller.start()

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            conn, addr = pair
            SMTPChannel(self, conn, addr)

    def process_message(self, Peer, From, To, Data, **kwargs):
        """
        Called by smtpd.SMTPServer when there's a message received.
        """

        try:
            logging.debug("Message received from Peer: %r, From: %r, to To %r.", Peer, From, To)
            routing.Router.deliver(mail.MailRequest(Peer, From, To, Data))
        except SMTPError as err:
            # looks like they want to return an error, so send it out
            return str(err)
        except Exception:
            logging.exception("Exception while processing message from Peer: %r, From: %r, to To %r.",
                              Peer, From, To)
            undeliverable_message(Data, "Error in message %r:%r:%r, look in logs." % (Peer, From, To))

    def close(self):
        """Doesn't do anything except log who called this, since nobody should.  Ever."""
        if six.PY3:
            trace = traceback.format_exc(chain=False)
        else:
            trace = traceback.format_exc()
        logging.error(trace)


class LMTPReceiver(lmtpd.LMTPServer):
    """Receives emails and hands it to the Router for further processing."""

    def __init__(self, host='127.0.0.1', port=8824, socket=None):
        """
        Initializes to bind on the given port and host/IP address. Remember that
        LMTP isn't for use over a WAN, so bind it to either a LAN address or
        localhost. If socket is not None, it will be assumed to be a path name
        and a UNIX socket will be set up instead.

        This uses lmtpd.LMTPServer in the __init__, which means that you have to
        call this far after you use python-daemonize or else daemonize will
        close the socket.
        """
        if socket is None:
            self.socket = "%s:%d" % (host, port)
            lmtpd.LMTPServer.__init__(self, (host, port))
        else:
            self.socket = socket
            lmtpd.LMTPServer.__init__(self, socket)

    def start(self):
        """
        Kicks everything into gear and starts listening on the port.  This
        fires off threads and waits until they are done.
        """
        logging.info("LMTPReceiver started on %s.", self.socket)
        self.poller = threading.Thread(target=asyncore.loop, kwargs={'timeout': 0.1, 'use_poll': True})
        self.poller.start()

    def process_message(self, Peer, From, To, Data, **kwargs):
        """
        Called by lmtpd.LMTPServer when there's a message received.
        """

        try:
            logging.debug("Message received from Peer: %r, From: %r, to To %r.", Peer, From, To)
            routing.Router.deliver(mail.MailRequest(Peer, From, To, Data))
        except SMTPError as err:
            # looks like they want to return an error, so send it out
            # and yes, you should still use SMTPError in your handlers
            return str(err)
        except Exception:
            logging.exception("Exception while processing message from Peer: %r, From: %r, to To %r.",
                              Peer, From, To)
            undeliverable_message(Data, "Error in message %r:%r:%r, look in logs." % (Peer, From, To))

    def close(self):
        """Doesn't do anything except log who called this, since nobody should.  Ever."""
        if six.PY3:
            trace = traceback.format_exc(chain=False)
        else:
            trace = traceback.format_exc()
        logging.error(trace)


class QueueReceiver(object):
    """
    Rather than listen on a socket this will watch a queue directory and
    process messages it receives from that.  It works in almost the exact
    same way otherwise.
    """

    def __init__(self, queue_dir, sleep=10, size_limit=0, oversize_dir=None):
        """
        The router should be fully configured and ready to work, the
        queue_dir can be a fully qualified path or relative.
        """
        self.queue = queue.Queue(queue_dir, pop_limit=size_limit,
                                 oversize_dir=oversize_dir)
        self.queue_dir = queue_dir
        self.sleep = sleep

    def start(self, one_shot=False):
        """
        Start simply loops indefinitely sleeping and pulling messages
        off for processing when they are available.

        If you give one_shot=True it will run once rather than do a big
        while loop with a sleep.
        """

        logging.info("Queue receiver started on queue dir %s", self.queue_dir)
        logging.debug("Sleeping for %d seconds...", self.sleep)

        inq = queue.Queue(self.queue_dir)

        while True:
            keys = inq.keys()

            for key in keys:
                msg = inq.get(key)

                if msg:
                    logging.debug("Pulled message with key: %r off", key)
                    self.process_message(msg)
                    logging.debug("Removed %r key from queue.", key)
                    inq.remove(key)

            if one_shot:
                return
            else:
                time.sleep(self.sleep)

    def process_message(self, msg):
        """
        Exactly the same as SMTPReceiver.process_message but just designed for the queue's
        quirks.
        """

        try:
            logging.debug("Message received from Peer: %r, From: %r, to To %r.", msg.Peer, msg.From, msg.To)
            routing.Router.deliver(msg)
        except SMTPError as err:
            # looks like they want to return an error, so send it out
            logging.exception("Raising SMTPError when running in a QueueReceiver is unsupported.")
            undeliverable_message(msg.Data, err.message)
        except Exception:
            logging.exception("Exception while processing message from Peer: "
                              "%r, From: %r, to To %r.", msg.Peer, msg.From, msg.To)
            undeliverable_message(msg.Data, "Router failed to catch exception.")



from nose.tools import *  # noqa
from salmon.testing import RouterConversation, queue, relay

relay = relay(port=8823)
client = RouterConversation("somedude@localhost", "requests_tests")
confirm_format = "testing-confirm-[0-9]+@"
noreply_format = "testing-noreply@"


def test_forwards_relay_host():
    """
    !!!!!! YOU MUST CONFIGURE YOUR config/settings.py OR THIS WILL FAIL !!!!!!
    Makes sure that your config/settings.py is configured to forward mail from
    localhost (or your direct host) to your relay.
    """
    client.begin()
    client.say("tester@localhost", "Test that forward works.", "tester@localhost")


def test_drops_open_relay_messages():
    """
    But, make sure that mail NOT for test.com gets dropped silently.
    """
    client.begin()
    client.say("tester@badplace.notinterwebs", "Relay should not happen")
    assert queue().count() == 0, "You are configured currently to accept everything. " \
                                 "You should change config/settings.py router_defaults " \
                                 "so host is your actual host name that will receive mail."



from mock import Mock, patch

from nose.tools import assert_equal
from salmon import mail
from salmon.confirm import ConfirmationEngine, ConfirmationStorage
from salmon.queue import Queue
from salmon.testing import delivered, relay

from .setup_env import setup_salmon_dirs, teardown_salmon_dirs

# quiten the linter on these variables
storage = None
engine = None


def setup_module(module):
    setup_salmon_dirs()
    module.storage = ConfirmationStorage()
    module.engine = ConfirmationEngine('run/confirm', storage)


def teardown_module(module):
    teardown_salmon_dirs()


def test_ConfirmationStorage():
    storage.store('testing', 'somedude@localhost',
                  '12345', '567890')
    secret, pending_id = storage.get('testing', 'somedude@localhost')
    assert_equal(secret, '12345')
    assert_equal(pending_id, '567890')

    storage.delete('testing', 'somedude@localhost')
    assert_equal(len(storage.confirmations), 0)

    storage.store('testing', 'somedude@localhost',
                  '12345', '567890')
    assert_equal(len(storage.confirmations), 1)
    storage.clear()
    assert_equal(len(storage.confirmations), 0)


@patch("smtplib.SMTP")
def test_ConfirmationEngine_send(smtp_mock):
    smtp_mock.return_value = Mock()

    Queue('run/queue').clear()
    engine.clear()

    list_name = 'testing'
    action = 'subscribing to'
    host = 'localhost'

    message = mail.MailRequest('fakepeer', 'somedude@localhost',
                               'testing-subscribe@localhost', 'Fake body.')

    engine.send(relay(port=0), 'testing', message, 'confirmation.msg', locals())

    assert_equal(smtp_mock.return_value.sendmail.call_count, 1)
    assert_equal(smtp_mock.return_value.quit.call_count, 1)
    assert delivered('somedude', to_queue=engine.pending)

    return smtp_mock.return_value.sendmail.call_args[0][2]


def test_ConfirmationEngine_verify():
    confirm = test_ConfirmationEngine_send()
    confirm = mail.MailRequest(None, None, None, confirm)

    resp = mail.MailRequest('fakepeer', '"Somedude Smith" <somedude@localhost>', confirm['Reply-To'], 'Fake body')

    target, _, expect_secret = confirm['Reply-To'].split('-')
    expect_secret = expect_secret.split('@')[0]

    found = engine.verify(target, resp['from'], 'invalid_secret')
    assert not found

    pending = engine.verify(target, resp['from'], expect_secret)
    assert pending, "Verify failed: %r not in %r." % (expect_secret,
                                                      storage.confirmations)

    assert_equal(pending['from'], 'somedude@localhost')
    assert_equal(pending['to'], 'testing-subscribe@localhost')


def test_ConfirmationEngine_cancel():
    confirm = test_ConfirmationEngine_send()
    confirm = mail.MailRequest(None, None, None, confirm)

    target, _, expect_secret = confirm['Reply-To'].split('-')
    expect_secret = expect_secret.split('@')[0]

    engine.cancel(target, confirm['To'], expect_secret)

    found = engine.verify(target, confirm['To'], expect_secret)
    assert not found



import logging
import logging.config

from salmon import queue
from salmon.routing import Router
from salmon.server import SMTPReceiver

from . import settings

logging.config.fileConfig("tests/config/logging.conf")

settings.receiver = SMTPReceiver(**settings.receiver_config)

Router.defaults(**settings.router_defaults)
Router.load(settings.dump_handlers)
Router.RELOAD = True
Router.UNDELIVERABLE_QUEUE = queue.Queue(settings.UNDELIVERABLE_QUEUE)



from mock import Mock, patch

from nose.tools import assert_equal, with_setup
from salmon.testing import RouterConversation, assert_in_state, clear_queue, delivered, queue, relay

from .setup_env import setup_salmon_dirs, teardown_salmon_dirs

relay = relay(port=0)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_clear_queue():
    queue().push("Test")
    assert_equal(queue().count(), 1)

    clear_queue()
    assert_equal(queue().count(), 0)


@patch("smtplib.SMTP")
@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_relay(smtp_mock):
    smtp_mock.return_value = Mock()

    relay.send('test@localhost', 'zedshaw@localhost', 'Test message', 'Test body')

    assert_equal(smtp_mock.return_value.sendmail.call_count, 1)
    assert_equal(smtp_mock.return_value.quit.call_count, 1)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_delivered():
    clear_queue()
    queue().push("To: gooduser@localhost\nFrom: tester@localhost\n\nHi\n")

    assert delivered("gooduser@localhost"), "Test message not delivered."
    assert not delivered("baduser@localhost")
    assert_in_state('salmon_tests.handlers.simple_fsm_mod', 'gooduser@localhost', 'tester@localhost', 'START')


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_RouterConversation():
    client = RouterConversation('tester@localhost', 'Test router conversations.')
    client.begin()
    client.say('testlist@localhost', 'This is a test')
    delivered('testlist@localhost'), "Test message not delivered."



from mock import patch

from nose.tools import assert_equal, with_setup
from salmon import utils

from .setup_env import setup_salmon_dirs, teardown_salmon_dirs


def clear_settings():
    utils.settings = None


@with_setup(teardown=clear_settings)
def test_make_fake_settings():
    settings = utils.make_fake_settings('localhost', 8800)
    assert settings
    assert settings.receiver
    assert settings.relay is None
    settings.receiver.close()


@with_setup(teardown=clear_settings)
def test_import_settings():
    assert utils.settings is None

    settings = utils.import_settings(True, boot_module='config.testing')
    assert settings
    assert settings.receiver_config
    assert_equal(settings, utils.settings)

    with patch("salmon.utils.importlib.import_module") as import_mock:
        # just import settings module
        clear_settings()
        utils.import_settings(False)
        assert_equal(import_mock.call_count, 1)

        # import settings and boot
        clear_settings()
        import_mock.reset_mock()
        utils.import_settings(True)
        assert_equal(import_mock.call_count, 2)

        # settings has already been imported, return early
        import_mock.reset_mock()
        utils.import_settings(False)
        assert_equal(import_mock.call_count, 0)

        # boot module doesn't get cached on the module, so it import_module should be called
        import_mock.reset_mock()
        utils.import_settings(True)
        assert_equal(import_mock.call_count, 1)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('daemon.DaemonContext.open')
def test_daemonize_not_fully(dc_open):
    context = utils.daemonize("run/tests.pid", ".", False, False, do_open=False)
    assert context
    assert_equal(dc_open.call_count, 0)
    dc_open.reset_mock()

    context = utils.daemonize("run/tests.pid", ".", "/tmp", 0o002, do_open=True)
    assert context
    assert_equal(dc_open.call_count, 1)


@patch("daemon.daemon.change_process_owner")
def test_drop_priv(cpo):
    utils.drop_priv(100, 100)
    assert_equal(cpo.call_count, 1)



# -*- coding: utf-8 -*-
from email import encoders
from email.utils import parseaddr
import mailbox
import os

from mock import Mock, patch
import chardet

from nose.tools import assert_equal, raises, with_setup
from salmon import encoding

from .setup_env import setup_salmon_dirs, teardown_salmon_dirs

BAD_HEADERS = [
    '"\u8003\u53d6\u5206\u4eab" <Ernest.Beard@msa.hinet.net>',
    '"=?windows-1251?B?RXhxdWlzaXRlIFJlcGxpY2E=?="\n\t<wolfem@barnagreatlakes.com>',
    '=?iso-2022-jp?B?Zmlicm91c19mYXZvcmF0ZUB5YWhvby5jby5qcA==?=<fibrous_favorate@yahoo.co.jp>',

    '=?windows-1252?Q?Global_Leadership_in_HandCare_-_Consumer,\n\t_Professional_and_Industrial_Products_OTC_:_FLKI?=',
    '=?windows-1252?q?Global_Leadership_in_Handcare_-_Consumer, _Auto,\n\t_Professional_&_Industrial_Products_-_OTC_:_FLKI?=',  # noqa
    'I am just normal.',
    '=?koi8-r?B?WW91ciBtYW6ScyBzdGFtaW5hIHdpbGwgY29tZSBiYWNrIHRvIHlvdSBs?=\n\t=?koi8-r?B?aWtlIGEgYm9vbWVyYW5nLg==?=',
    '=?koi8-r?B?WW91IGNhbiBiZSBvbiB0b3AgaW4gYmVkcm9vbSBhZ2FpbiCWIGp1c3Qg?=\n\t=?koi8-r?B?YXNrIHVzIGZvciBhZHZpY2Uu?=',
    '"=?koi8-r?B?5MXMz9DSz8na18/E09TXzw==?=" <daniel@specelec.com>',
    '=?utf-8?b?IumrlOiCsuWckuWNgOermSDihpIg6ZW35bqa6Yar6Zmi56uZIOKGkiDmlofljJbk?=\n =?utf-8?b?uInot6/nq5kiIDx2Z3hkcmp5Y2lAZG5zLmh0Lm5ldC50dz4=?=',  # noqa: E501
    '=?iso-8859-1?B?SOlhdnkgTel05WwgVW7uY/hk?=\n\t=?iso-8859-1?Q?=E9?=',
]

DECODED_HEADERS = encoding.header_from_mime_encoding(BAD_HEADERS)

NORMALIZED_HEADERS = [encoding.header_to_mime_encoding(x) for x in DECODED_HEADERS]


def test_MailBase():
    the_subject = u'p\xf6stal'
    m = encoding.MailBase()

    m['To'] = "testing@localhost"
    m['Subject'] = the_subject
    m['Content-Type'] = 'text/plain; charset=iso-8859-1'
    m['MIME-Version'] = '1.0'

    assert_equal(m['To'], "testing@localhost")
    assert_equal(m['TO'], m['To'])
    assert_equal(m['to'], m['To'])

    assert_equal(m['Subject'], the_subject)
    assert_equal(m['subject'], m['Subject'])
    assert_equal(m['sUbjeCt'], m['Subject'])

    msg = encoding.to_message(m)
    m2 = encoding.from_message(msg)

    assert_equal(len(m), len(m2))

    assert_equal(m.items(), m2.items())

    for k in m:
        assert m[k] == m2[k], "%s: %r != %r" % (k, m[k], m2[k])

    assert("To" in m)
    assert("Bob" not in m)
    assert_equal(m["To"], "testing@localhost")
    m["To"] = "somebody@localhost"
    assert_equal(m["To"], "somebody@localhost")
    assert_equal(m.keys(), ['Subject', 'Mime-Version', 'Content-Type', 'To'])
    assert_equal(m.items(), [
        ('Subject', the_subject),
        ('Mime-Version', '1.0'),
        ('Content-Type', 'text/plain; charset="iso-8859-1"'),
        ('To', 'somebody@localhost')])
    assert_equal(m.get_all("Subject"), [the_subject])
    assert_equal(m.get_all("Bob"), [])

    # appending headers
    m.append_header("subject", "something else")
    assert_equal(m["Subject"], the_subject)
    assert_equal(m.keys(), ['Subject', 'Mime-Version', 'Content-Type', 'To', "Subject"])
    assert_equal(m.items(), [
        ('Subject', the_subject),
        ('Mime-Version', '1.0'),
        ('Content-Type', 'text/plain; charset="iso-8859-1"'),
        ('To', 'somebody@localhost'),
        ('Subject', "something else")])
    assert_equal(m.get_all("Subject"), [the_subject, "something else"])

    # reset subject header
    m["Subject"] = the_subject

    for key, item in zip(m.keys(), m.items()):
        assert_equal(key, item[0])
        assert key in m
        del m[key]
        assert key not in m
        assert key not in [i[0] for i in m.items()]


def test_normalized_headers():
    base = encoding.MailBase()

    # __setitem__ will normalize headers when they're added
    base["subJect"] = "test"
    assert_equal(base["Subject"], "test")
    assert_equal(base.mime_part["Subject"], "test")
    assert_equal(base.keys(), ["Subject"])
    assert_equal(base.mime_part.keys(), ["Subject"])

    # if the header wasn't added via __setitem__, it should still be returned
    # to Salmon users in a normalized form
    base.mime_part["fRom"] = "test@example.com"
    assert_equal(base["From"], "test@example.com")
    assert_equal(base.mime_part["fRom"], "test@example.com")
    assert_equal(base.keys(), ["Subject", "From"])
    assert_equal(base.mime_part.keys(), ["Subject", "fRom"])


def test_header_to_mime_encoding():
    for i, header in enumerate(DECODED_HEADERS):
        assert_equal(NORMALIZED_HEADERS[i], encoding.header_to_mime_encoding(header))


def test_mutt_badness():
    # this is a sample of possibly the worst case Mutt can produce
    idiot = '=?iso-8859-1?B?SOlhdnkgTel05WwgVW7uY/hk?=\n\t=?iso-8859-1?Q?=E9?='
    should_be = u'H\xe9avy M\xe9t\xe5l Un\xeec\xf8d\xe9'
    assert_equal(encoding.header_from_mime_encoding(idiot), should_be)


def test_header_from_mime_encoding():
    assert not encoding.header_from_mime_encoding(None)
    assert_equal(len(BAD_HEADERS), len(encoding.header_from_mime_encoding(BAD_HEADERS)))

    for i, header in enumerate(BAD_HEADERS):
        assert_equal(DECODED_HEADERS[i], encoding.header_from_mime_encoding(header))


def test_to_message_from_message_with_spam():
    mb = mailbox.mbox("tests/spam")

    for msg in mb:
        m = encoding.from_message(msg)
        out = encoding.to_message(m)
        assert repr(out)

        m2 = encoding.from_message(out)

        for k in m:
            if '@' in m[k]:
                assert_equal(parseaddr(m[k]), parseaddr(m2[k]))
            elif k.lower() in [key.lower() for key in encoding.CONTENT_ENCODING_KEYS]:
                pass  # skip!
            else:
                assert m[k].strip() == m2[k].strip(), "%s: %r != %r" % (k, m[k], m2[k])

            assert not m[k].startswith(u"=?")
            assert not m2[k].startswith(u"=?")

        # salmon adds stuff to content headers that's missing from example messages
        for k in encoding.CONTENT_ENCODING_KEYS:
            if k == "Content-Transfer-Encoding" \
                    and m.content_encoding[k][0] is None \
                    and m2.content_encoding[k][0] == "7bit":
                continue  # salmon fixing bad email
            elif k == "Mime-Version" \
                    and m.content_encoding[k][0] is None \
                    and m2.content_encoding[k][0] == "1.0":
                continue  # again, salmon fixing bad email

            assert m.content_encoding[k][0] == m2.content_encoding[k][0], \
                "%s: %r != %r" % (k, m.content_encoding[k], m2.content_encoding[k])

            for p in m.content_encoding[k][1]:
                if p in m2.content_encoding[k][1]:
                    assert m.content_encoding[k][1][p].lower() == m2.content_encoding[k][1][p].lower(), \
                        "%s: %r != %r" % (p, m.content_encoding[k][1][p], m2.content_encoding[k][1][p])

        assert m.body == m2.body, "Bodies don't match"

        assert_equal(len(m.parts), len(m2.parts), "Not the same number of parts.")

        for i, part in enumerate(m.parts):
            assert part.body == m2.parts[i].body, \
                "Part %d isn't the same: %r \nvs\n. %r" % (i, part.body, m2.parts[i].body)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_to_file_from_file():
    mb = mailbox.mbox("tests/spam")
    msg = encoding.from_message(mb[0])

    outfile = "run/encoding_test.msg"

    with open(outfile, 'w') as outfp:
        encoding.to_file(msg, outfp)

    with open(outfile) as outfp:
        msg2 = encoding.from_file(outfp)

    assert_equal(len(msg), len(msg2))
    os.unlink(outfile)


def test_guess_encoding_and_decode():
    for header in DECODED_HEADERS:
        try:
            encoding.guess_encoding_and_decode('ascii', header.encode('utf-8'))
        except encoding.EncodingError:
            pass


def test_attempt_decoding():
    for header in DECODED_HEADERS:
        encoding.attempt_decoding('ascii', header.encode('utf-8'))


def test_properly_decode_header():
    for i, header in enumerate(BAD_HEADERS):
        parsed = encoding.properly_decode_header(header)
        assert_equal(DECODED_HEADERS[i], parsed)


def test_headers_round_trip():
    # round trip the headers to make sure they convert reliably back and forth
    for header in BAD_HEADERS:
        original = encoding.header_from_mime_encoding(header)

        assert original
        assert "=?" not in original and "?=" not in original, "Didn't decode: %r" % header

        encoded = encoding.header_to_mime_encoding(original)
        assert encoded

        return_original = encoding.header_from_mime_encoding(encoded)
        assert_equal(original, return_original)

        return_encoded = encoding.header_to_mime_encoding(return_original)
        assert_equal(encoded, return_encoded)


def test_MIMEPart():
    text1 = encoding.MIMEPart("text/plain")
    text1.set_payload("The first payload.")
    text2 = encoding.MIMEPart("text/plain")
    text2.set_payload("The second payload.")

    image_data = open("tests/salmon.png", "rb").read()
    img1 = encoding.MIMEPart("image/png")
    img1.set_payload(image_data)
    img1.set_param('attachment', '', header='Content-Disposition')
    img1.set_param('filename', 'salmon.png', header='Content-Disposition')
    encoders.encode_base64(img1)

    multi = encoding.MIMEPart("multipart/mixed")
    for x in [text1, text2, img1]:
        multi.attach(x)

    mail = encoding.from_message(multi)

    assert_equal(mail.parts[0].body, "The first payload.")
    assert_equal(mail.parts[1].body, "The second payload.")
    assert_equal(mail.parts[2].body, image_data)

    encoding.to_message(mail)


@patch('chardet.detect', new=Mock())
@raises(encoding.EncodingError)
def test_guess_encoding_fails_completely():
    chardet.detect.return_value = {'encoding': None, 'confidence': 0.0}
    encoding.guess_encoding_and_decode('ascii', 'some data', errors='strict')


def test_attach_text():
    mail = encoding.MailBase()
    mail.attach_text("This is some text.", 'text/plain')

    msg = encoding.to_message(mail)
    assert_equal(msg.get_payload(0).get_payload(), "This is some text.")
    assert encoding.to_string(mail)

    mail.attach_text("<html><body><p>Hi there.</p></body></html>", "text/html")
    msg = encoding.to_message(mail)
    assert_equal(len(msg.get_payload()), 2)
    assert encoding.to_string(mail)


def test_attach_file():
    mail = encoding.MailBase()
    png = open("tests/salmon.png", "rb").read()
    mail.attach_file("salmon.png", png, "image/png", "attachment")
    msg = encoding.to_message(mail)

    payload = msg.get_payload(0)
    assert_equal(payload.get_payload(decode=True), png)
    assert_equal(payload.get_filename(), "salmon.png")


def test_content_encoding_headers_are_maintained():
    inmail = encoding.from_file(open("tests/signed.msg", "rb"))

    ctype, ctype_params = inmail.content_encoding['Content-Type']

    assert_equal(ctype, 'multipart/signed')

    # these have to be maintained
    for key in ['protocol', 'micalg']:
        assert key in ctype_params

    # these get removed
    for key in encoding.CONTENT_ENCODING_REMOVED_PARAMS:
        assert key not in ctype_params

    outmsg = encoding.to_message(inmail)
    ctype, ctype_params = encoding.parse_parameter_header(outmsg, 'Content-Type')
    for key in ['protocol', 'micalg']:
        assert key in ctype_params, key


def test_odd_content_type_with_charset():
    mail = encoding.MailBase()
    mail.body = u"p\xf6stal".encode('utf-8')
    mail.content_encoding['Content-Type'] = ('application/plain', {'charset': 'utf-8'})

    msg = encoding.to_string(mail)
    assert msg


def test_specially_borked_lua_message():
    msg = encoding.from_file(open("tests/borked.msg", "rb"))
    assert msg


def raises_TypeError(*args):
    raise TypeError()


@patch('salmon.encoding.MIMEPart.__init__')
@raises(encoding.EncodingError)
def test_to_message_encoding_error(mp_init):
    mp_init.side_effect = raises_TypeError
    test = encoding.from_file(open("tests/borked.msg", "rb"))
    encoding.to_message(test)


@patch('salmon.encoding.chardet.detect')
@raises(encoding.EncodingError)
def test_guess_encoding_and_decode_unicode_error(detect_mock):
    data = u"testß".encode("latin-1")
    detect_mock.return_value = {"confidence": 0.99, "encoding": "utf-8"}
    encoding.guess_encoding_and_decode("ascii", data)


def test_attempt_decoding_with_bad_encoding_name():
    assert_equal("test", encoding.attempt_decoding("asdfasdf", "test"))


@raises(encoding.EncodingError)
def test_apply_charset_to_header_with_bad_encoding_char():
    encoding.apply_charset_to_header('ascii', 'X', 'bad')


def test_odd_roundtrip_bug():
    decoded_addrs = [
            u'"\u0414\u0435\u043b\u043e\u043f\u0440\u043e\u0438\u0437\u0432\u043e\u0434\u0441\u0442\u0432\u043e" <daniel@specelec.com>',  # noqa: E501
        u'"\u8003\u53d6\u5206\u4eab" <Ernest.Beard@msa.hinet.net>',
        u'"Exquisite Replica"\n\t<wolfem@barnagreatlakes.com>',
    ]

    for decoded in decoded_addrs:
        encoded = encoding.header_to_mime_encoding(decoded)
        assert '<' in encoded and '"' in encoded, "Address wasn't encoded correctly:\n%s" % encoded


def test_multiple_headers():
    msg = encoding.MailBase()
    msg["To"] = "somedude@localhost"
    msg["From"] = "nobody@localhost"
    assert("From" in msg)
    assert_equal(msg["From"], "nobody@localhost")
    msg["From"] = "somebody@localhost"
    assert_equal(msg["From"], "somebody@localhost")
    assert_equal(msg.keys(), ["To", "From"])

    # appending headers
    msg.append_header("To", "nobody@example.com")
    assert_equal(msg["To"], "somedude@localhost")
    assert_equal(msg.keys(), ["To", "From", "To"])



import mailbox
import os
import shutil

from mock import Mock, patch

from nose.tools import assert_equal, raises, with_setup
from salmon import mail, queue

from .setup_env import setup_salmon_dirs, teardown_salmon_dirs

USE_SAFE = False


BYTES_MESSAGE = u"""From: me@localhost
To: you@localhost
Subject: bob!

Blobcat""".encode()


def setup():
    if os.path.exists("run/big_queue"):
        shutil.rmtree("run/big_queue")


def teardown():
    setup()


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_push():
    q = queue.Queue("run/queue", safe=USE_SAFE)
    q.clear()

    # the queue doesn't really care if its a request or response, as long
    # as the object answers to str(msg)
    msg = mail.MailResponse(To="test@localhost", From="test@localhost", Subject="Test", Body="Test")
    key = q.push(msg)

    assert key, "Didn't get a key for test_get push."

    return q


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_pop():
    q = test_push()
    key, msg = q.pop()

    assert key, "Didn't get a key for test_get push."
    assert msg, "Didn't get a message for key %r" % key

    assert hasattr(msg, "Data"), "MailRequest doesn't have Data attribute"

    assert_equal(msg['to'], "test@localhost")
    assert_equal(msg['from'], "test@localhost")
    assert_equal(msg['subject'], "Test")
    assert_equal(msg.body(), "Test")

    assert_equal(q.count(), 0)
    assert not q.pop()[0]


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_get():
    q = test_push()
    msg = mail.MailResponse(To="test@localhost", From="test@localhost", Subject="Test", Body="Test")

    key = q.push(str(msg))
    assert key, "Didn't get a key for test_get push."

    msg = q.get(key)
    assert msg, "Didn't get a message for key %r" % key


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_remove():
    q = test_push()
    msg = mail.MailResponse(To="test@localhost", From="test@localhost", Subject="Test", Body="Test")

    key = q.push(str(msg))
    assert key, "Didn't get a key for test_get push."
    assert_equal(q.count(), 2)

    q.remove(key)
    assert_equal(q.count(), 1)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_safe_maildir():
    global USE_SAFE
    USE_SAFE = True
    test_push()
    test_pop()
    test_get()
    test_remove()


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_oversize_protections():
    # first just make an oversize limited queue
    overq = queue.Queue("run/queue", pop_limit=10)
    overq.clear()

    for i in range(5):
        overq.push("HELLO" * 100)

    assert_equal(overq.count(), 5)

    key, msg = overq.pop()

    assert not key and not msg, "Should get no messages."
    assert_equal(overq.count(), 0)

    # now make sure that oversize mail is moved to the overq
    setup()
    overq = queue.Queue("run/queue", pop_limit=10, oversize_dir="run/big_queue")
    moveq = queue.Queue("run/big_queue")

    for i in range(5):
        overq.push("HELLO" * 100)

    key, msg = overq.pop()

    assert not key and not msg, "Should get no messages."
    assert_equal(overq.count(), 0)
    assert_equal(moveq.count(), 5)

    moveq.clear()
    overq.clear()


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('os.stat', new=Mock())
@raises(mailbox.ExternalClashError)
def test_SafeMaildir_name_clash():
    sq = queue.SafeMaildir('run/queue')
    sq.add("TEST")


def raise_OSError(*x, **kw):
    err = OSError('Fail')
    err.errno = 0
    raise err


@patch('mailbox._create_carefully', new=Mock())
@raises(OSError)
def test_SafeMaildir_throws_errno_failure():
    setup()
    mailbox._create_carefully.side_effect = raise_OSError
    sq = queue.SafeMaildir('run/queue')
    sq.add("TEST")


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('os.stat', new=Mock())
@raises(OSError)
def test_SafeMaildir_reraise_weird_errno():
    os.stat.side_effect = raise_OSError
    sq = queue.SafeMaildir('run/queue')
    sq.add('TEST')


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_bytes():
    """Test that passing a queue raw data works, i.e. as happens in the
    undeliverable queue"""
    q = queue.Queue("run/queue", safe=USE_SAFE)
    q.clear()

    key = q.push(BYTES_MESSAGE)
    assert key, "Didn't get a key"

    mail = q.get(key)

    assert mail is not None, "Failed to get email from queue"

    assert_equal(mail['from'], "me@localhost")
    assert_equal(mail['to'], "you@localhost")
    assert_equal(mail['subject'], "bob!")
    assert_equal(mail.body(), "Blobcat")



import os
import shutil
import sys

from mock import Mock, patch

from nose.tools import assert_equal, assert_not_equal, raises, with_setup
from salmon import commands, encoding, mail, routing, utils

from .setup_env import setup_salmon_dirs, teardown_salmon_dirs


def setup():
    if os.path.exists("run/fake.pid"):
        os.unlink("run/fake.pid")


def teardown():
    if os.path.exists("run/fake.pid"):
        os.unlink("run/fake.pid")


def make_fake_pid_file():
    f = open("run/fake.pid", "w")
    f.write("0")
    f.close()


def get_command(command):
    def func(*args):
        args = list(args)
        args.insert(0, command)

        parsed_args = commands._parser.parse_args(args)

        cmd = parsed_args.func
        del parsed_args.func

        cmd(**vars(parsed_args))

    return func


@patch("salmon.server.smtplib.SMTP")
def test_send_command(client_mock):
    command = get_command("send")
    command("--sender", 'test@localhost', "--to", 'test@localhost', "--body",
            'Test body', "--subject", 'Test subject', "--attach", 'setup.py', "--port",
            "8899", "--host", "127.0.0.1")
    assert_equal(client_mock.return_value.sendmail.call_count, 1)


def test_status_command():
    command = get_command("status")
    command("--pid", 'run/log.pid')
    command("--pid", 'run/donotexist.pid')


@patch('sys.argv', ['salmon'])
@raises(SystemExit)
def test_main():
    commands.main()


@patch('salmon.queue.Queue')
@patch('sys.exit', new=Mock())
def test_queue_command(MockQueue):
    mq = MockQueue()
    mq.get.return_value = "A sample message"
    mq.keys.return_value = ["key1", "key2"]
    mq.pop.return_value = ('key1', 'message1')
    mq.count.return_value = 1

    command = get_command("queue")

    command("--pop")
    assert_equal(mq.pop.call_count, 1)

    command("--get", 'somekey')
    assert_equal(mq.get.call_count, 1)

    command("--remove", 'somekey')
    assert_equal(mq.remove.call_count, 1)

    command("--clear")
    assert_equal(mq.clear.call_count, 1)

    command("--keys")
    assert_equal(mq.keys.call_count, 1)

    command("--count")
    assert_equal(mq.count.call_count, 1)


@patch('sys.exit', new=Mock())
def test_gen_command():
    command = get_command("gen")
    project = 'tests/testproject'
    if os.path.exists(project):
        shutil.rmtree(project)

    command(project)
    assert os.path.exists(project)

    # test that it exits if the project exists
    command(project)
    assert_equal(sys.exit.call_args, ((1,),))

    sys.exit.reset_mock()
    command(project, "--force")
    assert_not_equal(sys.exit.call_args, ((1,),))

    # TODO: put this in a tear down
    shutil.rmtree(project)


def test_routes_command():
    command = get_command("routes")
    command('salmon.handlers.log', 'salmon.handlers.queue')

    # test with the --test option
    command('salmon.handlers.log', 'salmon.handlers.queue', "--test", "anything@localhost")

    # test with the -test option but no matches
    routing.Router.clear_routes()
    command("--test", "anything@localhost")


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('sys.exit', new=Mock())
@patch('salmon.utils.daemonize', new=Mock())
@patch('salmon.server.SMTPReceiver')
def test_log_command(MockSMTPReceiver):
    command = get_command("log")
    ms = MockSMTPReceiver()
    ms.start.function()

    setup()  # make sure it's clear for fake.pid
    command("--host", "127.0.0.1", "--port", "8825", "--pid", "run/fake.pid")
    assert_equal(utils.daemonize.call_count, 1)
    assert_equal(ms.start.call_count, 1)

    # test that it exits on existing pid
    make_fake_pid_file()
    command("--host", "127.0.0.1", "--port", "8825", "--pid", "run/fake.pid")
    assert_equal(sys.exit.call_count, 1)


@patch('sys.stdin', new=Mock())
@patch("salmon.server.smtplib.SMTP")
def test_sendmail_command(client_mock):
    sys.stdin.read.function()

    msg = mail.MailResponse(To="tests@localhost", From="tests@localhost",
                            Subject="Hello", Body="Test body.")
    sys.stdin.read.return_value = str(msg)

    command = get_command("sendmail")
    command("--host", "127.0.0.1", "--port", "8899", "test@localhost")
    assert_equal(client_mock.return_value.sendmail.call_count, 1)


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('sys.exit', new=Mock())
@patch('salmon.utils.daemonize', new=Mock())
@patch('salmon.utils.import_settings', new=Mock())
@patch('salmon.utils.drop_priv', new=Mock())
@patch('sys.path', new=Mock())
def test_start_command():
    # normal start
    command = get_command("start")
    command("--pid", "smtp.pid")
    assert_equal(utils.daemonize.call_count, 1)
    assert_equal(utils.import_settings.call_count, 1)

    # start with pid file existing already
    make_fake_pid_file()
    command("--pid", "run/fake.pid")
    assert_equal(sys.exit.call_count, 1)

    # start with pid file existing and force given
    assert os.path.exists("run/fake.pid")
    command("--force", "--pid", "run/fake.pid")
    assert not os.path.exists("run/fake.pid")

    # start with a uid but no gid
    command("--uid", "1000", "--pid", "run/fake.pid", "--force")
    assert_equal(utils.drop_priv.call_count, 0)

    # start with a uid/gid given that's valid
    command("--uid", "1000", "--gid", "1000", "--pid", "run/fake.pid", "--force")
    assert_equal(utils.drop_priv.call_count, 1)

    # non daemon start
    daemonize_call_count = utils.daemonize.call_count
    command("--pid", "run/fake.pid", "--no-daemon", "--force")
    assert_equal(utils.daemonize.call_count, daemonize_call_count)  # same count -> not called


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('sys.exit', new=Mock())
@patch('os.kill', new=Mock())
@patch('glob.glob', new=lambda x: ['run/fake.pid'])
def test_stop_command():
    command = get_command("stop")
    # gave a bad pid file
    command("--pid", "run/dontexit.pid")
    assert_equal(sys.exit.call_count, 1)

    make_fake_pid_file()
    command("--pid", "run/fake.pid")

    make_fake_pid_file()
    command("--all", "run")

    os_kill_count = os.kill.call_count
    make_fake_pid_file()
    command("--pid", "run/fake.pid", "--force")
    assert_equal(os.kill.call_count, os_kill_count + 1)  # kill should have been called once more
    assert not os.path.exists("run/fake.pid")

    make_fake_pid_file()
    os.kill.side_effect = OSError("Fail")
    command("--pid", "run/fake.pid", "--force")


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_cleanse_command():
    command = get_command("cleanse")
    command('run/queue', 'run/cleansed')
    assert os.path.exists('run/cleansed')


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
@patch('salmon.encoding.from_message')
def test_cleans_command_with_encoding_error(from_message):
    command = get_command("cleanse")
    from_message.side_effect = encoding.EncodingError
    command('run/queue', 'run/cleansed')


@with_setup(setup_salmon_dirs, teardown_salmon_dirs)
def test_blast_command():
    command = get_command("blast")
    command("--host", "127.0.0.1", "--port", "8899", "run/queue")



"""
A bag of generally useful things when writing unit tests for your Salmon
server.  The most important thing  using the TestConversation vs.
RouterConversation to talk to your server.

The TestConversation will use the salmon.server.Relay you have configured to
talk to your actual running Salmon server.  Since by default Salmon reloads
each file you change it will work to run your tests.

However, this isn't that fast, doesn't give you coverage analysis, and doesn't
let you test the results.  For that you use RouterConversation to do the exact
same API (they should be interchangeable) but rather than talk to a running
server through the relay, it just runs all the messages through the router
directly.

This is faster and will give you code coverage as well as make sure that all
the modules (not just your handlers) will get reloaded.
"""
from __future__ import print_function, unicode_literals

import re

from salmon import mail, routing, server
from salmon.queue import Queue

TEST_QUEUE = "run/queue"


def relay(hostname="127.0.0.1", port=8824):
    """Wires up a default relay on port 8824 (the default salmon log port)."""
    return server.Relay(hostname, port, debug=0)


def queue(queue_dir=TEST_QUEUE):
    """Creates a queue for you to analyze the results of a send, uses the
    TEST_QUEUE setting in settings.py if that exists, otherwise defaults to
    run/queue."""
    return Queue(queue_dir)


def clear_queue(queue_dir=TEST_QUEUE):
    """Clears the default test queue out, as created by salmon.testing.queue."""
    queue(queue_dir).clear()


def delivered(pattern, to_queue=None):
    """
    Checks that a message with that patter is delivered, and then returns it.

    It does this by searching through the queue directory and finding anything that
    matches the pattern regex.
    """
    inq = to_queue or queue()
    for key in inq.keys():
        msg = inq.get(key)
        if not msg:
            # no messages in the queue
            return False

        regp = re.compile(pattern)
        if regp.search(pattern.__class__(msg)):
            msg = inq.get(key)
            return msg

    # didn't find anything
    return False


class TestConversation(object):
    """
    Used to easily do conversations with an email server such that you
    send a message and then expect certain responses.
    """

    def __init__(self, relay_to_use, From, Subject):
        """
        This creates a set of default values for the conversation so that you
        can easily send most basic message.  Each method lets you override the
        Subject and Body when you send.
        """
        self.relay = relay_to_use
        self.From = From
        self.Subject = Subject

    def begin(self):
        """Clears out the queue and Router states so that you have a fresh start."""
        clear_queue()
        routing.Router.clear_states()

    def deliver(self, To, From, Subject, Body):
        """Delivers it to the relay."""
        self.relay.send(To, From, Subject, Body)

    def say(self, To, Body, expect=None, Subject=None):
        """
        Say something to To and expect a reply with a certain address.
        It returns the message expected or None.
        """
        msg = None

        self.deliver(To, self.From, Subject or self.Subject, Body)
        if expect:
            msg = delivered(expect)
            if not msg:
                print("MESSAGE IN QUEUE:")
                inq = queue()
                for key in inq.keys():
                    print("-----")
                    print(inq.get(key))

            assert msg, "Expected %r when sending to %r with '%s:%s' message." % (expect,
                                                                                  To, self.Subject or Subject, Body)
        return msg


class RouterConversation(TestConversation):
    """
    An implementation of TestConversation that routes the messages
    internally to the Router, rather than connecting with a relay.
    Use it in tests that are not integration tests.
    """

    def __init__(self, From, Subject):
        self.From = From
        self.Subject = Subject

    def deliver(self, To, From, Subject, Body):
        """Overrides TestConversation.deliver to do it internally."""
        sample = mail.MailResponse(From=From, To=To, Subject=Subject, Body=Body)
        msg = mail.MailRequest('localhost', sample['From'], sample['To'], str(sample))
        routing.Router.deliver(msg)


def assert_in_state(module, To, From, state):
    """
    Makes sure a user is in a certain state for a certain user.
    Use these sparingly, since every time you change your handler you'll
    have to change up your tests.  It's better to focus on the interaction
    with your handler and expected outputs.
    """
    fake = {'to': To}
    state_key = routing.Router.state_key(module, fake)
    assert routing.Router.STATE_STORE.get(state_key, From) == state, \
        "%r != %r" % (routing.Router.STATE_STORE.get(state_key, From), state)



# This file contains python variables that configure Salmon for email processing.

# You may add additional parameters such as `username' and `password' if your
# relay server requires authentication, `starttls' (boolean) or `ssl' (boolean)
# for secure connections.
relay_config = {'host': 'localhost', 'port': 8825}

receiver_config = {'host': 'localhost', 'port': 8823}

handlers = ['app.handlers.sample']

router_defaults = {'host': '.+'}

# the config/boot.py will turn these values into variables set in settings



from __future__ import print_function

from shutil import rmtree
import os

from salmon.routing import Router

from . import dirs


def clear_dir(dir_path):
    for dir_item in os.listdir(dir_path):
        print(dir_item)
        full_path = '%s/%s' % (dir_path, dir_item)
        print(full_path)
        if os.path.isfile(full_path):
            os.unlink(full_path)
        else:
            rmtree(full_path)


def setup_salmon_dirs():
    for path in dirs:
        clear_dir(path)


def teardown_salmon_dirs():
    for path in dirs:
        clear_dir(path)


def setup_router(handlers):
    Router.clear_routes()
    Router.clear_states()
    Router.HANDLERS.clear()
    Router.load(handlers)



"""
Bounce analysis module for Salmon.  It uses an algorithm that tries
to simply collect the headers that are most likely found in a bounce
message, and then determine a probability based on what it finds.
"""
from __future__ import print_function, unicode_literals

from functools import wraps
import re

BOUNCE_MATCHERS = {
    'Action': re.compile(r'(failed|delayed|delivered|relayed|expanded)', re.IGNORECASE | re.DOTALL),
    'Content-Description': re.compile(r'(Notification|Undelivered Message|Delivery Report)', re.IGNORECASE | re.DOTALL),
    'Diagnostic-Code': re.compile(r'(.+);\s*([0-9\-\.]+)?\s*(.*)', re.IGNORECASE | re.DOTALL),
    'Final-Recipient': re.compile(r'(.+);\s*(.*)', re.IGNORECASE | re.DOTALL),
    'Received': re.compile(r'(.+)', re.IGNORECASE | re.DOTALL),
    'Remote-Mta': re.compile(r'(.+);\s*(.*)', re.IGNORECASE | re.DOTALL),
    'Reporting-Mta': re.compile(r'(.+);\s*(.*)', re.IGNORECASE | re.DOTALL),
    'Status': re.compile(r'([0-9]+)\.([0-9]+)\.([0-9]+)', re.IGNORECASE | re.DOTALL)
}

BOUNCE_MAX = len(BOUNCE_MATCHERS) * 2.0

PRIMARY_STATUS_CODES = {
    u'1': u'Unknown Status Code 1',
    u'2': u'Success',
    u'3': u'Temporary Failure',
    u'4': u'Persistent Transient Failure',
    u'5': u'Permanent Failure'
}

SECONDARY_STATUS_CODES = {
    u'0':   u'Other or Undefined Status',
    u'1':   u'Addressing Status',
    u'2':   u'Mailbox Status',
    u'3':   u'Mail System Status',
    u'4':   u'Network and Routing Status',
    u'5':   u'Mail Delivery Protocol Status',
    u'6':   u'Message Content or Media Status',
    u'7':   u'Security or Policy Status',
}

COMBINED_STATUS_CODES = {
    u'00': u'Not Applicable',
    u'10': u'Other address status',
    u'11': u'Bad destination mailbox address',
    u'12': u'Bad destination system address',
    u'13': u'Bad destination mailbox address syntax',
    u'14': u'Destination mailbox address ambiguous',
    u'15': u'Destination mailbox address valid',
    u'16': u'Mailbox has moved',
    u'17': u'Bad sender\'s mailbox address syntax',
    u'18': u'Bad sender\'s system address',

    u'20': u'Other or undefined mailbox status',
    u'21': u'Mailbox disabled, not accepting messages',
    u'22': u'Mailbox full',
    u'23': u'Message length exceeds administrative limit.',
    u'24': u'Mailing list expansion problem',

    u'30': u'Other or undefined mail system status',
    u'31': u'Mail system full',
    u'32': u'System not accepting network messages',
    u'33': u'System not capable of selected features',
    u'34': u'Message too big for system',

    u'40': u'Other or undefined network or routing status',
    u'41': u'No answer from host',
    u'42': u'Bad connection',
    u'43': u'Routing server failure',
    u'44': u'Unable to route',
    u'45': u'Network congestion',
    u'46': u'Routing loop detected',
    u'47': u'Delivery time expired',

    u'50': u'Other or undefined protocol status',
    u'51': u'Invalid command',
    u'52': u'Syntax error',
    u'53': u'Too many recipients',
    u'54': u'Invalid command arguments',
    u'55': u'Wrong protocol version',

    u'60': u'Other or undefined media error',
    u'61': u'Media not supported',
    u'62': u'Conversion required and prohibited',
    u'63': u'Conversion required but not supported',
    u'64': u'Conversion with loss performed',
    u'65': u'Conversion failed',

    u'70': u'Other or undefined security status',
    u'71': u'Delivery not authorized, message refused',
    u'72': u'Mailing list expansion prohibited',
    u'73': u'Security conversion required but not possible',
    u'74': u'Security features not supported',
    u'75': u'Cryptographic failure',
    u'76': u'Cryptographic algorithm not supported',
    u'77': u'Message integrity failure',
}


def match_bounce_headers(msg):
    """
    Goes through the headers in a potential bounce message recursively
    and collects all the answers for the usual bounce headers.
    """
    matches = {'Content-Description-Parts': {}}
    for part in msg.base.walk():
        for k in BOUNCE_MATCHERS:
            if k in part:
                if k not in matches:
                    matches[k] = set()

                # kind of an odd place to put this, but it's the easiest way
                if k == 'Content-Description':
                    matches['Content-Description-Parts'][part[k].lower()] = part

                matches[k].add(part[k])

    return matches


def detect(msg):
    """
    Given a message, this will calculate a probability score based on
    possible bounce headers it finds and return a salmon.bounce.BounceAnalyzer
    object for further analysis.

    The detection algorithm is very simple but still accurate.  For each header
    it finds it adds a point to the score.  It then uses the regex in BOUNCE_MATCHERS
    to see if the value of that header is parsable, and if it is it adds another
    point to the score.  The final probability is based on how many headers and matchers
    were found out of the total possible.

    Finally, a header will be included in the score if it doesn't match in value, but
    it WILL NOT be included in the headers used by BounceAnalyzer to give you meanings
    like remote_mta and such.

    Because this algorithm is very dumb, you are free to add to BOUNCE_MATCHERS in your
    boot files if there's special headers you need to detect in your own code.
    """
    originals = match_bounce_headers(msg)
    results = {'Content-Description-Parts':
               originals['Content-Description-Parts']}
    score = 0
    del originals['Content-Description-Parts']

    for key in originals:
        score += 1  # score still goes up, even if value doesn't parse
        r = BOUNCE_MATCHERS[key]

        scan = (r.match(v) for v in originals[key])
        matched = [m.groups() for m in scan if m]

        # a key is counted in the score, but only added if it matches
        if len(matched) > 0:
            score += len(matched) / len(originals[key])
            results[key] = matched

    return BounceAnalyzer(results, score / BOUNCE_MAX)


class BounceAnalyzer(object):
    """
    BounceAnalyzer collects up the score and the headers and gives more
    meaningful interaction with them.  You can keep it simple and just use
    is_hard, is_soft, and probable methods to see if there was a bounce.
    If you need more information then attributes are set for each of the following:

        * primary_status -- The main status number that determines hard vs soft.
        * secondary_status -- Advice status.
        * combined_status -- the 2nd and 3rd number combined gives more detail.
        * remote_mta -- The MTA that you sent mail to and aborted.
        * reporting_mta -- The MTA that was sending the mail and has to report to you.
        * diagnostic_codes -- Human readable codes usually with info from the provider.
        * action -- Usually 'failed', and turns out to be not too useful.
        * content_parts -- All the attachments found as a hash keyed by the type.
        * original -- The original message, if it's found.
        * report -- All report elements, as salmon.encoding.MailBase raw messages.
        * notification -- Usually the detailed reason you bounced.
    """
    def __init__(self, headers, score):
        """
        Initializes all the various attributes you can use to analyze the bounce
        results.
        """
        self.headers = headers
        self.score = score

        if 'Status' in self.headers:
            status = self.headers['Status'][0]
            self.primary_status = int(status[0]), PRIMARY_STATUS_CODES[status[0]]
            self.secondary_status = int(status[1]), SECONDARY_STATUS_CODES[status[1]]
            combined = "".join(status[1:])
            self.combined_status = int(combined), COMBINED_STATUS_CODES[combined]
        else:
            self.primary_status = (None, None)
            self.secondary_status = (None, None)
            self.combined_status = (None, None)

        if 'Remote-Mta' in self.headers:
            self.remote_mta = self.headers['Remote-Mta'][0][1]
        else:
            self.remote_mta = None

        if 'Reporting-Mta' in self.headers:
            self.reporting_mta = self.headers['Reporting-Mta'][0][1]
        else:
            self.reporting_mta = None

        if 'Final-Recipient' in self.headers:
            self.final_recipient = self.headers['Final-Recipient'][0][1]
        else:
            self.final_recipient = None

        if 'Diagnostic-Code' in self.headers:
            self.diagnostic_codes = self.headers['Diagnostic-Code'][0][1:]
        else:
            self.diagnostic_codes = [None, None]

        if 'Action' in self.headers:
            self.action = self.headers['Action'][0][0]
        else:
            self.action = None

        # these are forced lowercase because they're so damn random
        self.content_parts = self.headers['Content-Description-Parts']
        # and of course, this isn't the original original, it's the wrapper
        self.original = self.content_parts.get('undelivered message', None)

        if self.original and self.original.parts:
            self.original = self.original.parts[0]

        self.report = self.content_parts.get('delivery report', None)
        if self.report and self.report.parts:
            self.report = self.report.parts

        self.notification = self.content_parts.get('notification', None)

    def is_hard(self):
        """
        Tells you if this was a hard bounce, which is determined by the message
        being a probably bounce with a primary_status greater than 4.
        """
        return self.probable() and self.primary_status[0] > 4

    def is_soft(self):
        """Basically the inverse of is_hard()"""
        return self.probable() and self.primary_status[0] <= 4

    def probable(self, threshold=0.3):
        """
        Determines if this is probably a bounce based on the score
        probability.  Default threshold is 0.3 which is conservative.
        """
        return self.score > threshold

    def error_for_humans(self):
        """
        Constructs an error from the status codes that you can print to
        a user.
        """
        if self.primary_status[0]:
            return "%s, %s, %s" % (self.primary_status[1],
                                   self.secondary_status[1],
                                   self.combined_status[1])
        else:
            return "No status codes found in bounce message."


class bounce_to(object):
    """
    Used to route bounce messages to a handler for either soft or hard bounces.
    Set the soft/hard parameters to the function that represents the handler.
    The function should take one argument of the message that it needs to handle
    and should have a route that handles everything.

    WARNING: You should only place this on the START of modules that will
    receive bounces, and every bounce handler should return START.  The reason
    is that the bounce emails come from *mail daemons* not the actual person
    who bounced.  You can find out who that person is using
    message.bounce.final_recipient.  But the bounce handler is *actually*
    interacting with a message from something like MAILER-DAEMON@somehost.com.
    If you don't go back to start immediately then you will mess with the state
    for this address, which can be bad.
    """
    def __init__(self, soft=None, hard=None):
        self.soft = soft
        self.hard = hard

        assert self.soft and self.hard, "You must give at least soft and/or hard"

    def __call__(self, func):
        @wraps(func)
        def bounce_wrapper(message, *args, **kw):
            if message.is_bounce():
                if message.bounce.is_soft():
                    return self.soft(message)
                else:
                    return self.hard(message)
            else:
                return func(message, *args, **kw)

        return bounce_wrapper



"""
Implements a forwarding handler that will take anything it receives and
forwards it to the relay host.  It is intended to use with the
salmon.routing.RoutingBase.UNDELIVERABLE_QUEUE if you want mail that Salmon
doesn't understand to be delivered like normal.  The Router will dump
any mail that doesn't match into that queue if you set it, and then you can
load this handler into a special queue receiver to have it forwarded on.

BE VERY CAREFUL WITH THIS.  It should only be used in testing scenarios as
it can turn your server into an open relay if you're not careful.  You
are probably better off writing your own version of this that knows a list
of allowed hosts your machine answers to and only forwards those.
"""

import logging

from salmon.routing import route, stateless
from salmon.utils import settings


@route("(to)@(host)", to=".+", host=".+")
@stateless
def START(message, to=None, host=None):
    """Forwards every mail it gets to the relay.  BE CAREFUL WITH THIS."""
    logging.debug("MESSAGE to %s@%s forwarded to the relay host.", to, host)
    settings.relay.deliver(message)



from mock import Mock, patch

from nose.tools import assert_equal, with_setup
from salmon import mail, utils
from salmon.routing import Router

sample_message = """From: someone@localhost
To: someone@localhost

Test
"""


def create_message():
    return mail.MailRequest("localhost", "someone@localhost", "someone@localhost", sample_message)


def cleanup_router():
    Router.clear_routes()
    Router.clear_states()
    Router.HANDLERS.clear()
    utils.settings = None


@with_setup(teardown=cleanup_router)
def test_log_handler():
    import salmon.handlers.log  # noqa
    Router.deliver(create_message())


@with_setup(teardown=cleanup_router)
def test_queue_handler():
    import salmon.handlers.queue  # noqa
    Router.deliver(create_message())


@patch("smtplib.SMTP")
@with_setup(teardown=cleanup_router)
def test_forward(smtp_mock):
    smtp_mock.return_value = Mock()

    utils.import_settings(False)

    import salmon.handlers.forward  # noqa
    Router.deliver(create_message())

    assert_equal(smtp_mock.return_value.sendmail.call_count, 1)
    assert_equal(smtp_mock.return_value.quit.call_count, 1)
